# -*- coding: utf-8 -*-
"""
E"-binning single-path nonuniform temperature fitting based on time-domain fitting.

Splits H2O linelist into several separate databases for bins of lower-state energy.
Then performs multi-molecule time-domain fit (each E"-bin treated as different molecule)

Syntax warning: uses object-oriented programming with inheritance.

Abbreviations:
    td = time-domain signal. inverse Fourier Transform of frequency spectrum
    wvn = wavenumber (cm-1) of x-axis
    HTP = Hartman-Tran lineshape profile
    bl = laser transmission baseline. number of time-domain points attributed to baseline
    e = lower-state energy (cm-1)
    ebin = subset of linelist for H2O for transitions with similar lower-state energy
    s = linestrength
    rat = ratio, multiplied by some normalization value (reference pressure, reference linestrength) to get actual value
    sh = s_hat = normalized linestrength
    srat = integrated_area / (S(T_ref)P_ref X_ref L_ref)
    TPXL = path-average thermodynamic parameters
            t_celsius (temperature Celsius), 
            p_torr (pressure Torr), 
            chi (molefraction of absorbing molecule (default H2O) (between 0-1)), 
            pathlength (cm)

Created on Tue Jul 23 09:00:06 2019

@author: Nate the Average
"""
# built-in modules
import os
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from time import time
import json
from scipy.optimize import leastsq, nnls # for least-squares fitting
#   use scipy instead of lmfit to handle arbitrary number of linelist inputs

# support modules in the package
try:
    from .Ebin_support import snorm_fit_t_x, s_two_temp, s_t_exact
    from .Ebin_support import calc_cepstrum_td, calc_spectrum_fd
    from .td_support import weight_func
    from .file_support import Transmission
    from ..hapi import db_begin, HITRAN_DEFAULT_HEADER
except:
    from Ebin_support import snorm_fit_t_x, s_two_temp, s_t_exact
    from Ebin_support import calc_cepstrum_td, calc_spectrum_fd
    from td_support import weight_func
    from file_support import Transmission
    from hapi import db_begin, HITRAN_DEFAULT_HEADER


## external module for calculating model absorption spectra.
#from packfind import find_package
#find_package('pldspectrapy')
#import pldspectrapy.pldhapi as hapi

class EbinHapi(Transmission):
    """
    Object-oriented script for time-domain E"-binning. 
    Designed for H2O-in-air
    
    INPUTS:
        output_path  = output directory + name of output files.
           eg 'C:\\Users\\Nate\\Documents\\fits\\H2O_5Feb' will make Labfit files called 'H2O_5Feb.inp'
           in the directory 'C:\\Users\\Nate\\Documents\\fits
        bandwidth     = (cm-1) fitting window--should be contained within transmission spectrum.
        
    OUTPUTS:
        self.snorm_out (nonuniform linestrength curve calculated against reference TPXL)
        (information also saved in output_path.txt for export to TikhonovInversion)
        
        Other optional outputs:
        - other spectral fit plots available through spectrum_by_ebin(), plot_ebin()
        - coarse temperature fits available through fit_temperature() and temperature_bin()
        - absorption spectrum for nonuniform path calculated through simulate_spectrum()

        
    Created on Wed Oct 31 15:52:38 2018    
    @author: Nate the Average
    """
    
    def __init__(self, output_path, bandwidth = [6800,7100], Labfit_HTP = False):
        # Load generic Labfit variables
        if bandwidth[0] > bandwidth[1]:
            foo = bandwidth[1]
            bandwidth[1] = bandwidth[0]
            bandwidth[0] = foo
        Transmission.__init__(self, bandwidth)
        # And initialize E"-binning-specific variables
        self.Elist = [0,500,1000,2000,7000]#[0,50,100,200,300,400,500,600,750,1020,1220,1440,1680,1940,2220,2510,2820,3100,4000,7000]
            # A list to hit all of the J" = 5-16 family of high-E" lines. vib = (101<-000)
            # E"=1000-3000 bins split along middle between each strong-line J" family.
            # Any lines with E" > max(self.Elist) will not be included in the fit
        self.bl = 0.002 # baseline weighting portion
        self.etalons = []
        self.weight = None
        self.snorm_out = []
        self.p_rat = None # If model features narrower than measurement, increase p_rat
        self.shift = None # If v_model < v_meas, then reduce shift (more negative)
        self.sdvoigt = False
        
        # Initialize the spectral variables
        self.bkgd_mod_fd = None
        self.flip_spectrum = False # for DCS if reverse Nyquist window

        # optional true discretized T(x) profile
        self.tx_act_kelvin = None
        
        # Initialize linelist variables
        if os.path.isabs(output_path) is True:
            self.full_output_name = output_path
        else:
            self.full_output_name = os.path.join(os.getcwd(), output_path)
        self.path, self.output_name = os.path.split(output_path)
        self.path = os.path.join(self.path, 'hood')
        if os.path.exists(self.path) is False:
            os.makedirs(self.path)
        self.linelist_path = ''
        self.wvn_start = bandwidth[0]
        self.wvn_stop  = bandwidth[1] 
        self.n_lines = 0
        self.molec_id = 1
        self.local_iso_id = 1
        
    def calc_bkgd(self, p_atm = 0, molefraction = 0, t_kelvin = 296,
                  pathlength_cm = 0, shift_wvn = 0, db_override=''):
        '''
        Calculate time-domain background-subtraction model from P,x,T,L
        '''
        assert self.x_wvn is not None, "Need absorption spectrum first"
        
        # Set up linelist file
        if len(db_override) == 0:
            db_override = self.linelist_path
        db_name = os.path.basename(db_override).split('.')[0]
        dir_db = os.path.dirname(db_override)
        if len(dir_db) == 0:
            db_begin(os.getcwd())
        else:
            db_begin(dir_db)
            
        # Setup T,P,x,L parameters
        self.bkgd_pars = {'p_atm':p_atm}
        self.bkgd_pars['chi'] = molefraction
        self.bkgd_pars['t_kelvin'] = t_kelvin
        self.bkgd_pars['pathlength_cm'] = pathlength_cm
        self.bkgd_pars['shift_wvn'] = shift_wvn

        # Simulate bkgd
        self.bkgd_mod_fd = calc_spectrum_fd(self.x_wvn, 1, 1, molefraction,
                                           p_atm, t_kelvin,
                                           pathlength_cm, shift_wvn, 
                                           db_name, self.sdvoigt)

###############################################
# Nest functions used in fit_snorm_* functions #
###############################################           
    def ebin(self,x):
        '''
        Determine which E"-bin a line with E" = x is in. Returns index of Elist. 
        
        Nested function of make_srat()
        '''
#        for i in range(1,len(self.Elist)):
#            if(x < self.Elist[i] and x >= self.Elist[i-1]):
#                return i-1
        try:
            ebin_i = next(index for index, e_val in enumerate(self.Elist) if e_val > x)
        except StopIteration:
            ebin_i = len(self.Elist)
        return ebin_i - 1
            
    def prep_ebins(self):
        '''
        Make different hapi file for each ebin
        '''
        msg = "Provide a valid linelist file path for self.linelist_path"
        assert os.path.exists(self.linelist_path), msg
        header = get_linelist_header(self.linelist_path)

        nE = len(self.Elist)-1 # first row never assigned.
        Ecount_weight = np.zeros(nE)
        Esum = np.zeros(nE)
        self.par_name = []
        par_bin = []
        if len(self.snorm_out) != nE:
            # Initialize normalized linestrengths if changed E"-bins,
            #  otherwise keep the output from the previous fit (or file load)
            self.snorm_out = np.ones(nE)
            self.n_lines = 0
        try:
            [os.remove(os.path.join(self.path,f)) for f in os.listdir(self.path) 
                                       if self.output_name+'_' in f and '.data' in f]
            [os.remove(os.path.join(self.path,f)) for f in os.listdir(self.path) 
                               if self.output_name+'_' in f and '.header' in f]
        except FileNotFoundError:
            pass
        for i in range(nE):
            e_i = int(1/2 * (self.Elist[i] + self.Elist[i+1]))
            self.par_name.append(self.output_name + '_' + repr(e_i))
#            self.par_name.append(self.full_output_name + '_' + repr(e_i) + '.data')
            par_bin.append(open(os.path.join(self.path,self.par_name[-1]+'.data'),'w'))
            # and write out the header file
            with open(os.path.join(self.path, self.par_name[-1] + '.header'),'w') as f:
                f.write(json.dumps(header, indent=2))
        
        # Now sort each line into different E"-bin data file
        with open(self.linelist_path, 'r') as lines:
            for line in lines:
                self.molec_id = int(line[:2])
                self.local_iso_id = int(line[2])
                nu = float(line[3:15])
                if nu > self.wvn_stop:
                    break
                if nu > self.wvn_start:
                    elower = float(line[45:55])
                    ebin_i = self.ebin(elower)
                    s_i = float(line[15:25])* s_t_exact(self.t_celsius+273, 296, elower)
                    if ebin_i < nE and ebin_i >= 0:
                        Ecount_weight[ebin_i] += s_i**2
                        Esum[ebin_i] += s_i**2 * elower 
                        par_bin[ebin_i].write(line)
                        self.n_lines += 1
        for ebin in par_bin:
            ebin.close()
        par_bin = []
        # Weighted-average E" for each bin for future temperature inversion
        self.ebin_xVals = np.round(Esum / Ecount_weight)
        # Check on E"-selections
        self.bin_size = Ecount_weight # bigger number is better for fit uncertainty
        if min(self.bin_size) == 0:
            print('WARNING: certain E"-bins have no absorption features')
            bad = 0
            bads = []
            for i, b in enumerate(self.bin_size):
                if b == 0:
                    print('E" = ' + repr(self.Elist[i]) + '-' + repr(self.Elist[i+1]))
                    # Delete empty E"-bins from Elist and linelist files
                    bad_file = os.path.join(self.path, self.par_name[i])
                    os.remove(bad_file + '.data')
                    os.remove(bad_file + '.header')
                    bad += 1
                    bads.append(i)
            # Delete empty E"-bins from all relevant files
            self.ebin_xVals = np.delete(self.ebin_xVals, bads)
            self.Elist = np.delete(self.Elist, np.array(bads)+1)
            self.par_name = np.delete(self.par_name, bads)
            if isinstance(self.shift, np.ndarray): # also check size of self.shift!
                self.shift = np.ones(len(self.par_name) - bad) * np.average(self.shift)
                
        return Ecount_weight
    
    def get_td(self):
        warning = "Import arrays for self.trans_spectrum and self.x_wvn, and "
        warning += "check whether self.print_thermo() is making decent average-PxTL."
        assert self.trans_spectrum is not None, warning
        
        self.absorption_td = np.fft.irfft(-np.log(self.trans_spectrum))
        self.absorption_td[0] = 0; self.absorption_td[-1] = 0 # remove dc
        if self.bl < 1: # only have fraction of time-domain to unweight
            self.bl = int(self.bl*len(self.absorption_td))
        self.weight = weight_func(len(self.trans_spectrum), self.bl, self.etalons)
        if self.bkgd_mod_fd is not None:
            self.absorption_td -= np.fft.irfft(self.bkgd_mod_fd)
            
    def shift_setup(self):
        '''
        Rescale dimension of width and shift arrays.
        Avoids runtime error with E"-bin model calculations
        '''
        nbins = len(self.Elist) - 1
        if self.p_rat is None:
            self.p_rat = np.ones(nbins)
        elif len(self.p_rat) != nbins:
            self.p_rat = np.average(self.p_rat) * np.ones(nbins)
        if self.shift is None:
            self.shift = np.zeros(nbins)
        if isinstance(self.shift, float) or isinstance(self.shift, int):
            self.shift = self.shift * np.ones(nbins)
        if len(self.shift) != nbins:
            self.shift = np.average(self.shift) * np.ones(nbins)
            
    def spectrum_by_ebin(self):
        '''
        Given fit results, plot absorbance for each spectrum.
        Must run some fit_snorm_* first
        '''
        if self.x_wvn is None:
            print('WARNING: no loaded transmission spectrum exists')
            print('Setting nominal transmission = 1, 0.01cm-1 resolution')
            self.x_wvn = np.arange(self.wvn_start, self.wvn_stop, 0.01)
            self.trans_spectrum = np.ones(len(self.x_wvn))
        self.absorption_models = []
        self.absorption_total = np.zeros(len(self.x_wvn))
        self.shift_setup()
        for ix, ebin in enumerate(self.par_name):
            p_atm = self.p_torr / 760 * self.p_rat[ix]
            absorption_ebin = calc_spectrum_fd(self.x_wvn, self.molec_id, 
                                               self.local_iso_id, self.chi,
                                               p_atm, self.t_celsius+273,
                                               self.pathlength, self.shift[ix], 
                                               ebin, self.sdvoigt)
            absorption_ebin *= self.snorm_out[ix] / self.p_rat[ix]
            self.absorption_total += absorption_ebin
            self.absorption_models.append(absorption_ebin)
        fig, axs = plt.subplots(2,1, sharex='col', figsize=(9,4))
        for index, ebin in enumerate(self.absorption_models):
            axs[0].plot(self.x_wvn, ebin, label = 'E" = ' + repr(self.ebin_xVals[index]))
        axs[1].set_xlabel('Wavenumber (cm$^{-1}$)'); axs[0].set_ylabel('Absorbance'); 
#        axs[0].legend(loc='upper left')
        try:
            res = self.absorption_td - np.fft.irfft(self.absorption_total)
        except: # AttributeError or ValueError
            self.get_td()
            res = self.absorption_td - np.fft.irfft(self.absorption_total)
        self.bl_mod = np.real(np.fft.rfft((1-self.weight)*res))
        self.data_meas = np.real(np.fft.rfft(self.absorption_td) - self.bl_mod)
        axs[0].plot(self.x_wvn, self.data_meas, 'k:', label = 'Data')
        axs[1].plot(self.x_wvn, self.data_meas - self.absorption_total,'r', linewidth=0.8)
        axs[1].set_xlim(self.wvn_start, self.wvn_stop)
        axs[0].set_title(os.path.basename(self.full_output_name))

###############################################################################
# Analysis functions you might call directly after running fit_snorm_* functions
###############################################################################
    def plot_ebin(self, ebin_index, tight_plot=False, plot_trans = False):
        '''
        After running spectrum_by_ebin, compare fit for single bin of features.
        Show srat uncertainty in shaded error bars,
        show all other absorption contributions in blue
        '''
        absorption_ebin = self.absorption_models[ebin_index]
        absorption_other = self.absorption_total - absorption_ebin
        uc_rat = self.snorm_uc[ebin_index] / self.snorm_out[ebin_index]
        colors = {'ebin':'r','others':'xkcd:bright blue'}
        if ebin_index >= 0:
            elabel='%d<E"<%d bin' % (self.Elist[ebin_index], self.Elist[ebin_index+1])
        else:
            elabel='%d<E"<%d bin' % (self.Elist[ebin_index-1], self.Elist[ebin_index])
#        # plot linestrength curve
#        plt.figure()
#        plt.plot(self.ebin_xVals, self.snorm_out,'.',color=colors['others'])
#        plt.errorbar(self.ebin_xVals, self.snorm_out, self.snorm_uc, color=colors['others'], ls='none')
#        plt.plot(self.ebin_xVals[ebin_index], self.snorm_out[ebin_index],'.',color=colors['ebin'])
#        plt.errorbar(self.ebin_xVals[ebin_index], self.snorm_out[ebin_index], 
#                     self.snorm_uc[ebin_index], color=colors['ebin'], ls='none')
#        plt.xlabel('E"'); plt.ylabel('Normalized linestrength')
        if tight_plot:
            # Plot without horizontal space between axes,
            #  tough for zooming but good for output plots
            plt.rcParams.update({'xtick.direction':'in','xtick.top':True})
            gs1 = gridspec.GridSpec(3,1)
            gs1.update(hspace=0)
            fig = plt.figure();
            # transmission spectrum
            if plot_trans:
                ax1 = plt.subplot(gs1[0])
                ax1.plot(self.x_wvn, self.trans_spectrum, 'k', linewidth=0.8)
                ax1.set_xlim(self.wvn_start, self.wvn_stop)
                ax1.set_xticklabels([])
                ax1.set_ylabel('Transmission')
                ax1.set_title('Time-domain fit %d < E" < %d' % (self.Elist[ebin_index], self.Elist[ebin_index+1]))
            else:
                ax1 = plt.subplot(3,1,1)
                ax1.plot(self.ebin_xVals, self.snorm_out,'.',
                         markersize=3, color=colors['others'])
                ax1.errorbar(self.ebin_xVals, self.snorm_out, self.snorm_uc,
                             color=colors['others'], ls='none')
                ax1.plot(self.ebin_xVals[ebin_index], self.snorm_out[ebin_index],
                         '+',markersize = 7, color=colors['ebin'])
                ax1.errorbar(self.ebin_xVals[ebin_index], self.snorm_out[ebin_index], 
                             self.snorm_uc[ebin_index], color=colors['ebin'], ls='none')      
                ax1.set_xlabel('E"'); ax1.set_ylabel('S')
            # absorbance spectrum
            ax2 = plt.subplot(gs1[1])
            other, = ax2.plot(self.x_wvn, absorption_other, color=colors['others'], linewidth=0.5, label = 'All other E"-bins')
            meas, = ax2.plot(self.x_wvn, self.data_meas, 'k:', linewidth=1.1, label='Measurement')
            ax2.fill_between(self.x_wvn, (1-uc_rat) * absorption_ebin,
                         (1+uc_rat) * absorption_ebin, color='r', alpha=0.6)
            main, = ax2.plot(self.x_wvn, absorption_ebin,color=colors['ebin'], label = elabel) 
            handles = [meas, main, other]
            ax2.legend(handles,['Measurement',elabel,'All other E"-bins'],frameon=False)
            ax2.set_xlim(self.wvn_start, self.wvn_stop)
            ax2.set_xticklabels([])
            ax2.set_ylabel('Absorbance')
            # residual spectrum
            ax3 = plt.subplot(gs1[2])
            ax3.plot(self.x_wvn, self.data_meas - self.absorption_total,'k', linewidth=0.8)
            ax3.set_xlim(self.wvn_start, self.wvn_stop)
            ax3.set_xlabel('Wavenumber (cm$^{-1}$)')
            ax3.set_ylabel('Residual')
            return fig, ax1, ax2, ax3
        else:
            fig, axs = plt.subplots(2,1, sharex='col')
            other, = axs[0].plot(self.x_wvn, absorption_other,color=colors['others'], 
                        linewidth=.5, label = 'All other E"-bins')
            meas, = axs[0].plot(self.x_wvn, self.data_meas,'k:', linewidth=1.1, label = 'Measurement')
            # and plot uncertainty of E"-bin
            axs[0].fill_between(self.x_wvn, (1-uc_rat) * absorption_ebin,
                             (1+uc_rat) * absorption_ebin, color='r', alpha=0.6) 
            main, = axs[0].plot(self.x_wvn, absorption_ebin,color=colors['ebin'], label = elabel)       
            handles = [meas, main, other]
            # residual plot and formatting
            axs[1].plot(self.x_wvn, self.data_meas - self.absorption_total,'k', linewidth=0.8)
            axs[1].plot(self.x_wvn, np.zeros(len(self.x_wvn)), 'b:')
            axs[1].set_xlabel('Wavenumber (cm$^{-1}$)')
            axs[1].set_ylabel('Absorbance residual')
            axs[0].set_ylabel('Absorbance')
            axs[0].legend(handles,['Measurement',elabel,'All other E"-bins'])
            axs[0].set_title(self.par_name[ebin_index])
            axs[1].set_xlim(self.wvn_start, self.wvn_stop)
    
    def plot_td(self):
        '''
        See time-domain fit residual and weighting window.
        Must run spectrum_by_ebin() first
        '''
        plt.figure()
        plt.plot(self.absorption_td, label = 'data')
        res_td = self.absorption_td - np.fft.irfft(self.absorption_total)
        plt.plot(res_td, label = 'residual')
        plt.plot(self.weight, 'k', label = 'weighting window')
        plt.ylim(-.01, .01)
        plt.xlabel('Time')
        plt.ylabel('Absorbance')
    
    def fit_temperature(self, change_temperature = False, two_t_fit = False):
        '''
        Quick fits to S(E") curves (with plots)
        Scatter plot of individual S(E") and regression for path-average T,X fit
        
        INPUTS:
            change_temperature = can update T,X estimate from regression
                                 then iterate on fit_snorm()
                                 to fit with improved lineshape model
            two_t_fit = Calculate second trendline for 2-zone path
                        with T1, X1, T2, X2
                        If this fits data better, then evidence for nonuniformity
                        WARNING: finicky, unstable nonlinear fitting routine
        '''
        plt.figure(); 
        plt.plot(self.ebin_xVals, self.snorm_out, 'b.')
        plt.errorbar(self.ebin_xVals, self.snorm_out, self.snorm_uc,
                  label = 'spectral data fit', ls='none', color='b')
        t0 = self.t_celsius + 273
        elower = np.linspace(self.ebin_xVals[0], self.ebin_xVals[-1])
        self.ebin_output = np.zeros((len(self.snorm_out),3))
        self.ebin_output[:,0] = self.ebin_xVals
        self.ebin_output[:,1] = self.snorm_out
        self.ebin_output[:,2] = self.snorm_uc
        # single-temperature fit
        t_fit, x_fit = leastsq(snorm_fit_t_x, [self.t_celsius, self.chi], 
                         args = (self.ebin_output, [self.t_celsius, self.chi]))[0]
        print('Path-average temperature ~ %d Celsius (%d K)' % (t_fit, (273+t_fit)))
        print('Path-average molefraction ~ %.3f' % x_fit)
        s_mod = s_t_exact(t_fit+273, t0, elower) * (x_fit / self.chi)
        plt.plot(elower, s_mod, label = 'single-T fit')
        plt.xlabel('E"'); plt.ylabel('Normalized linestrength');
        plt.legend()
        if change_temperature:
            self.t_celsius = int(t_fit)
            self.chi = x_fit
            self.p_rat = None # overwrite default pressures
            
            if self.chi > 1:
                self.chi = 1
                print('Warning: Molefraction too strong.')
                print('Double-check pressure and pathlength.')
        if two_t_fit:
            # two-temperature fit
            t1, t2, x1, x2 = leastsq(s_two_temp, [300, self.t_celsius, .1, self.chi],
                                     args = (self.ebin_output, self.t_celsius, self.chi))[0]
            s_mod = 1 / self.chi * (x1 * s_t_exact(t1+273, t0+273, elower)
                                                + x2 * s_t_exact(t2+273, t0+273, elower))
            plt.plot(elower, s_mod, label = 'two-T fit')
            plt.legend()
            return t1, t2, x1, x2
    
    def temperature_bin(self, t_min = 300, t_max = 1300):
        '''
        Non-negative least-squares inversion solution
         for Temperature Distribution Function (See Part 1 paper, section 4.1)
        
        If you don't want to do full TikhonovInversion, this will give you 
         a coarse estimate of the distribution of H2O across temperature.

        Same algorithm as MATLAB's lsqnonneg(A,b) from Lawson, Hanson        
        '''
        n_t = (t_max - t_min) + 1
        t_i = np.arange(t_min, t_max+1)
        matrix = np.zeros((len(self.ebin_xVals), n_t))
        for t in range(n_t):
            matrix[:,t] = s_t_exact(t_min + t, self.t_celsius + 273, self.ebin_xVals)
        tdf_t = nnls(matrix, self.snorm_out)[0]
        plt.figure(); plt.errorbar(self.ebin_xVals, self.snorm_out, self.snorm_uc)
        plt.plot(self.ebin_xVals, matrix @ tdf_t)
        plt.xlabel('E"'); plt.ylabel('Normalized linestrength');
        plt.legend('data','fit')
        plt.figure(); plt.plot(t_i, tdf_t);
        t_avg = np.sum(t_i * tdf_t) / np.sum(tdf_t)
        print('Fit temperature', t_avg, 'K')
        plt.xlabel('Temperature (K)'); plt.ylabel('TDF')
        return tdf_t

    def ebin_select(self, low_e = 0, high_e = 4000, de = 100, uc_thrsh = 5):
        '''
        Iterative algorithm for selecting optimal E"-bins.
        
        Run fit_snorm() with a dense linear-spaced selection of E",
        then combine the E"-bins above some statistical uncertainty threshold.
        
        INPUTS:
            low_e = lowest included E" (cm-1)
            high_e = highest included E" (cm-1)
            de = width of each E"-bin (cm-1)
            uc_thrsh = aggregate E"-bins with uncertainty above this relative
                       snorm_uc threshold
       OUTPUT:
           no explicit outputs
           optimized self.Elist
           accompanying normalized linestrengths from area-fit only
        '''
        self.Elist = np.arange(low_e, high_e, de)
        self.fit_snorm(False);
        uc_thrsh *= min(self.snorm_uc) # scale threshold to lowest-UC E"-bin
        bads = np.where(self.snorm_uc > uc_thrsh)
#        bads = np.floor(self.snorm_uc / uc_thrsh).nonzero() # alternate method
        self.Elist = np.delete(self.Elist, np.array(bads)+1)
        self.fit_snorm();
        
    def ebin_select2(self, s_thrsh=1e-22, de_thrsh = 100):
        '''
        Select E"-bins based on two principles:
            1) Merge bins if there are stronger features at lower-E" AND higher-E"
            2) Extend total E" range as long as there are features above noise.
                
        This method works best if good self.t_celsius estimate.
        Can iterate on fit_snorm and fit_temperature 
        to determine t_celsius and noise floor threshold from snorm_uc
        TODO:
            figure out noise floor
            Account for nearby strong features
            Center the E" bins around strong features
            Figure out different rules of thumb for high-model error cases
            
        BUG: make sure si2 is not resized
        '''
        Elist = np.arange(0,7000,10)
        self.Elist = Elist.copy() # start with arbitrarily high # bins
        si2 = self.prep_ebins() # linestrength sum-of-squares in each E"-bin
        # si2 estimate 
        # 1) Check for stronger E"-bins at lower and higher E"
        strongest = np.argmax(si2)
        m = strongest
        ebins = [m]
        # 2) 
        while m > 0 and np.sqrt(sum(si2[:m])) > s_thrsh:
            # check if sandwiched between lower E"-bins
            low = np.argmax(si2[:m-1])
            if low < (m-1): # there are stronger E"-bins below
                de = Elist[m] - Elist[low]
                if de > de_thrsh:
                    # 2) Check for strength of E"-bins
                    s_bin = np.sqrt(sum(si2[low:m]))
                    if s_bin > s_thrsh:
                        # Add the E"-bin
                        ebins.append(low) # actually average around low
                        m = low
                else:
                    # Buggy algorithm here.
                    low2 = np.argmax(si2[:low-1])
                    ebins.append(low2)
                    m = low2
                    pass # reduce low
            else:
                m -= 1
                pass # I don't understand this case
        m = strongest
        while m < len(si2)-1 and np.sqrt(sum(si2[m:])) > s_thrsh:
            # check if sandwiched between lower E"-bins
            hi = m-1 + np.argmax(si2[m+1:])
            if hi > (m+1): # there are stronger E"-bins below
                try:
                    de = Elist[hi] - Elist[m]
                except IndexError:
                    print(ebins)
                    print('Trying to set %d out of %d' % (hi, len(self.Elist)))
                if de > de_thrsh:
                    # 2) Check for strength of E"-bins
                    s_bin = np.sqrt(sum(si2[m:hi]))
                    if s_bin > s_thrsh:
                        # Add the E"-bin
                        ebins.append(hi) # actually average around low
                        m = hi
                else:
                    # Buggy algorithm here.
                    hi2 = hi-1+np.argmax(si2[hi+1:])
                    ebins.append(hi2)
                    m = hi2
                    pass # reduce low
            else:
                m += 1
                pass # I don't understand this case
        ebins = np.sort(np.asarray(ebins))
        # Now pick the E"-bin boundaries halfway between the strong features
        evals = Elist[ebins]
        ebounds = np.zeros(len(evals))
        ebounds[1:] = .5 * evals[:-1] + .5 * evals[1:]
        
        self.Elist = ebounds            

################################################################
## Random function to calculate nonuniform absorption spectra ##
################################################################
    def simulate_spectra(self, calc_uniform = False):
        '''
        Simulate nonuniform frequency-domain absorption spectrum.
        '''
        assert self.tx_act_kelvin is not None, "Populate self.tx_act_kelvin with temperature profile array"
        msg = "Provide a valid linelist file path for self.linelist_path"
        assert os.path.exists(self.linelist_path), msg
        if self.x_wvn is None:
            print('WARNING: no loaded frequency axis exists')
            print('Setting nominal 0.01cm-1 resolution')
            self.x_wvn = np.arange(self.wvn_start, self.wvn_stop, 0.01)
        # Set up the source file database
        source_file_path = os.path.dirname(os.path.abspath(self.linelist_path))
        db_name = os.path.basename(self.linelist_path).split('.')[0]
        switch_paths = False
        if source_file_path != os.path.abspath(self.path):
            switch_paths = True
            db_begin(source_file_path)
        
        # add together each quasi-uniform chunk
        nchunks = len(self.tx_act_kelvin)
        t_avg = 0
        alpha_nonuni = np.zeros(len(self.x_wvn))
        for t in self.tx_act_kelvin:
            t_avg += t / nchunks
            alpha_nonuni += calc_spectrum_fd(self.x_wvn, self.molec_id, self.local_iso_id,
                                             self.chi, self.p_torr/760, t,
                                             self.pathlength/nchunks, 0, db_name, self.sdvoigt)
        plt.figure()
        plt.plot(self.x_wvn, alpha_nonuni, 'r', label = 'nonuniform T(x)')
        plt.xlabel('Wavenumber (cm$^{-1}$)')
        plt.ylabel('Absorbance')
        plt.title('Simulated spectra for tx_act_kelvin')
        alpha_uni = None
        if calc_uniform:
            alpha_uni = calc_spectrum_fd(self.x_wvn, self.molec_id, self.local_iso_id,
                                         self.chi, self.p_torr/760, t_avg,
                                         self.pathlength, 0, db_name, self.sdvoigt)
            plt.plot(self.x_wvn, alpha_uni, 'k--', label = 'uniform path-average T')
            plt.legend(frameon = False)
        # return to old linelist path
        if switch_paths:
            db_begin(self.path)
        return alpha_nonuni, alpha_uni

############################################################
## Spectral-fitting functions with different constraints. ##
############################################################     
    def fit_snorm(self, plot_fit = True, ebin_correlations = False):
        '''
        Swiftest, least-accurate normalized linestrength fit routine.
        Solves some linear scaling factor (self.snorm_out)
         for the absorption spectra of each E"-bin,
         holding lineshapes fixed at estimated path-average TPXL values.
         
        If path is not too nonuniform, and estimated TPXL come from previous 
         uniform-path fit, then this approximation often works well,
         especially for E" = 1500-3000 (cm-1)
        
        Uses linear-regression fitting of m-FID signal (time-domain)
        
        INPUT:
            plot_fit = plots all E"-bins, but increases run time
                        set to False for iterated fit_snorm(); fit_temperature()
        '''
        tic = time()
        self.prep_ebins()
        db_begin(self.path)
        self.get_td()
        # lineshape scaling
#        self.p_rat = None # scale by p_torr
        self.shift_setup()
        
        # calculate the spectra for each E"-bin
        #  (assume no shift) !!
        dalpha_ds = np.zeros(((len(self.x_wvn)-1) * 2,len(self.par_name)))
        for i, db_name in enumerate(self.par_name):
            a_td = calc_cepstrum_td(self.x_wvn, self.molec_id, self.local_iso_id,
                                    self.chi, self.p_torr / 760,
                                    self.t_celsius + 273, self.pathlength, 
                                    self.shift[i], db_name, self.sdvoigt)
            dalpha_ds[:,i] = a_td * self.weight

        # Now linear regression with chi2 uncertainty
        meas_w = self.absorption_td * self.weight
        cov_s = np.linalg.inv(np.transpose(dalpha_ds) @ dalpha_ds)
        self.snorm_out = cov_s @ np.transpose(dalpha_ds) @ (meas_w)
        res_w = meas_w - dalpha_ds @ self.snorm_out
        chi2 = np.sum(res_w**2) / (len(res_w) - len(self.snorm_out))
        self.snorm_uc = np.sqrt(np.diag(cov_s) * chi2)
        
        if ebin_correlations:
            # normalize covariance matrix to show correlations between E"-bins
            rho_s = cov_s.copy()
            for i in range(np.shape(cov_s)[0]):
                rho_s[:,i] /= np.sqrt(cov_s[i,i])
                rho_s[i,:] /= np.sqrt(cov_s[i,i])
                rho_s[i,i] = 0 # should be 1, but set to 0 to emphaize GUI color-code
            return rho_s

        # wrap-up operations
        self.ebin_output = np.transpose(np.asfarray([self.ebin_xVals, self.snorm_out, self.snorm_uc]))
        if plot_fit:
            self.spectrum_by_ebin()
            self.save_to_tikhonov()
        print('Spectral fit in %d seconds' % (time() - tic))
        return self.snorm_out

    def fit_snorm_width(self):
        '''
        Floats Lorentz width and column density for each E"-bin.
        Updates path-average pressure according to broadening.
        
        Fit takes longer than for fit_snorm, (it's nonlinear least-squares)
        but is more likely to converge than is fit_snorm_width_shift

        Uses scipy.optimize.leastsq for time-domain fitting.
            first input (spectra_float_width) is time-domain measurement
            second input (self.snorm_out) is fit parameters
            third input (args) are all other parameters to calculate time-domain absorption model
              which are held constant throughout the fit.
        
        '''
        tic = time()
        self.prep_ebins()
        db_begin(self.path)
        self.get_td()
        # lineshape scaling
        vars = list(self.snorm_out)
        nbins = len(self.par_name)
        p_torr_before = self.p_torr # Do I want to just pass this to shift_setup?
        self.shift_setup()
        vars.extend(list(p_torr_before/760 * self.p_rat))
        
        out = leastsq(spectra_float_width, vars, args=(self.x_wvn,
                                                            self.molec_id,
                                                            self.local_iso_id,
                                                            self.t_celsius+273,
                                                            self.chi,
                                                            self.pathlength,
                                                            self.absorption_td,
                                                            self.weight,
                                                            self.shift,
                                                            self.par_name, self.sdvoigt),
                                                        full_output = True)
        if out[1] is None:
            print("Fit did not converge")
        else:
            res = out[2]['fvec']
            cov_x = out[1] * np.sum(res**2) / (len(res) - len(out[0]))
            length_out = out[0][:nbins]
            p_out_atm = out[0][-nbins:]
            self.p_torr = np.average(p_out_atm) * 760
            self.p_rat =  p_out_atm / (self.p_torr/760)
            self.snorm_out = length_out * self.p_rat
            self.snorm_uc = np.sqrt(np.diag(cov_x))[:nbins]
            self.ebin_output = np.transpose(np.asfarray([self.ebin_xVals, self.snorm_out, self.snorm_uc]))
            self.spectrum_by_ebin()
            self.save_to_tikhonov()
            print('Spectral fit in %d seconds' % (time() - tic))
        return out
    
    def fit_snorm_width_shift(self, fix_bins = []):
        '''
        Floats Lorentz width, lineshift and column density for each E"-bin.
        
        CAUTION: shift-fits are very unstable with this LM fitting.
        This fit will probably not converge unless you already have good
        snorm estimates from fit_snorm().
        
        Pressure-shift and -broadening follow a (296/T) power law,
        so the low-E" lines can have different linewidths and linecenters
        than those calculated for path-average temperature.
        
        INPUT:
            fix_bins = indices of any E"-bins to not fit for width and shift
        '''
        tic = time()
        self.prep_ebins()
        db_begin(self.path)
        self.get_td()
        vars = list(self.snorm_out)
        # Setup lineshape fits
        self.shift_setup()
        nbins = len(self.par_name)
        n_fix = len(fix_bins)
        float_bins = np.delete(np.arange(len(self.shift)), fix_bins)
        p_atm = self.p_torr/760 * self.p_rat
        vars.extend(list(p_atm[float_bins]))
        vars.extend(list(self.shift[float_bins]))        
        # Scipy fit engine
        out = leastsq(spectra_float_width_shift, vars, args=(self.x_wvn,
                                                            self.molec_id,
                                                            self.local_iso_id,
                                                            self.t_celsius+273,
                                                            self.chi,
                                                            self.pathlength,
                                                            self.absorption_td,
                                                            self.weight,
                                                            self.par_name, self.sdvoigt,
                                                            fix_bins,
                                                            p_atm[fix_bins],
                                                            self.shift[fix_bins]),
                                                        full_output = True)
        if out[1] is None:
            print("Fit did not converge")
        else:
            res = out[2]['fvec']
            cov_x = out[1] * np.sum(res**2) / (len(res) - len(out[0]))
            length_out = out[0][:nbins]
            self.p_rat[float_bins] = out[0][nbins:2*nbins-n_fix] / (self.p_torr/760)
            self.shift[float_bins] = out[0][-(nbins-n_fix):]
            self.snorm_out = length_out * self.p_rat
            self.snorm_uc = np.sqrt(np.diag(cov_x))[:nbins]
            self.ebin_output = np.transpose(np.asfarray([self.ebin_xVals, self.snorm_out, self.snorm_uc]))
            self.spectrum_by_ebin()
            self.save_to_tikhonov()
#            self.p_torr = np.average(self.p_rat) * 760 
            print('Spectral fit in %d seconds' % (time() - tic))
        return out
        
#####################################
##  Data/file management functions ##
#####################################
    def print_thermo(self):
        print("Uniform-path approximations")
        print("Temperature = ", self.t_celsius, "Celsius")
        print("Pressure = ", self.p_torr, "Torr")
        print("Molefraction = ", self.chi)
        print("Pathlength = ", self.pathlength, "cm")
        print('\n')
        print('\n')
        print('To change these properties for spectral fitting, ')
        print('update self.t_celsius, self.p_torr, self.chi, self.pathlength.')
        print('eg. >> N.t_celsius = 500   # (if working with Example_td.py)')

    def save_fit(self, suffix = '', full_path = True):
        '''
        Save dictionary of relevant fit conditions for future Python load.
        Can then access from new Python command line using EbinHapi.load_fit()
        
        INPUTS:
            suffix = append to file name (so you can save multiple fits,
                    for instance one of fit_snorm, and another of fit_snorm_width)
            full_path = if True, save full file location of transmission + linelist files
                        set False if exporting fit to another computer,
                        to use relative file path within zipped directories
            
        '''
        
        # External file handling (linelist and transmission spectrum)

        if len(self.transmission_path) == 0:
            print('Do not know file location of transmission spectrum')
            print('Saving current trans_spectrum and x_wvn')
            self.make_asc(self.full_output_name)
        if full_path:
            fit = {'linelist_path':os.path.abspath(self.linelist_path)}
            fit['transmission_path'] = os.path.abspath(self.transmission_path)
        else:
            fit = {'linelist_path':os.path.relpath(self.linelist_path)}
            fit['transmission_path'] = os.path.relpath(self.transmission_path)
        
        # Variable and array handling
        fit['wvn_start'] = self.wvn_start
        fit['wvn_stop'] = self.wvn_stop
        fit['t_celsius'] = self.t_celsius
        fit['pressure'] = self.p_torr
        fit['molefraction'] = self.chi
        fit['pathlength'] = self.pathlength
        if isinstance(self.Elist, np.ndarray):
            self.Elist = self.Elist.astype('float')
        fit['Elist'] = list(self.Elist)
        fit['sdvoigt'] = self.sdvoigt
        fit['snorm_out'] = list(self.snorm_out)
        fit['snorm_uc'] = list(self.snorm_uc)
        fit['bl'] = self.bl
        fit['etalons'] = self.etalons
        try:
            fit['p_rat'] = list(self.p_rat)
        except:
            pass
        try:
            fit['shift'] = list(self.shift)
        except:
            pass

#        return fit
    
        with open(self.full_output_name + '_ebin' + suffix,'w') as f:
            f.write(json.dumps(fit, indent=2))
            
    def load_fit(self, suffix='_ebin',load_transmission = False, load_path_override = None):
        '''
        Load conditions from previous fit, saved under save_fit with self.full_output_name
        Should just initialize EbinHapi, run load_fit, and then be able to run plot_ebin()
        INPUTS:
            load_transmission = flag to also load transmission spectrum from .asc file
            load_path_override = file path of previous fit 
                                 (defaults to EbinHapi.full_output_name)
        '''
        message = 'Fitting '
        if load_path_override is None:
            with open(self.full_output_name + suffix,'r') as f:
                fit = json.loads(f.read())
        else:
            with open(load_path_override,'r') as f:
                fit = json.loads(f.read())
        self.wvn_start = fit['wvn_start']
        self.wvn_stop = fit['wvn_stop']
        try:
            self.t_celsius = fit['t_celsius']
        except: # old notation
            self.t_celsius = fit['temp_c']
        self.p_torr = fit['pressure']
        self.chi = fit['molefraction']
        self.pathlength = fit['pathlength']
        self.Elist = fit['Elist']
        self.snorm_out = np.asarray(fit['snorm_out'])
        self.snorm_uc = np.asarray(fit['snorm_uc'])
        message += repr(len(self.snorm_out)) + ' areas'
        self.bl = fit['bl']
        try:
            self.etalons = fit['etalons']
        except KeyError:
            pass
        try:
            self.p_rat = np.asarray(fit['p_rat'])
            message += ', broadening'
        except KeyError:
            pass     
        try:
            self.shift = np.asarray(fit['shift'])
            message += ', shifts'
        except KeyError:
            pass
        # and prepare linelist from previous time for plotting
        self.linelist_path = fit['linelist_path']
        try:
            self.sdvoigt = fit['sdvoigt']
        except:
            pass
        self.prep_ebins()
        db_begin(self.path)
        if 'bkgd_pars' in fit.keys():
            self.bkgd_pars = fit['bkgd_pars']
                
        message += '.'
        print(message)
        if load_transmission:
            asc_file_path = fit['transmission_path']
            self.get_transmission_asc(asc_file_path, overwrite_thermo=False)
            if 'bkgd_pars' in fit.keys():
                self.calc_bkgd(self.bkgd_pars['p_atm'],
                               self.bkgd_pars['chi'],
                               self.bkgd_pars['t_kelvin'],
                               self.bkgd_pars['pathlength_cm'],
                               self.bkgd_pars['shift_wvn'])
            self.spectrum_by_ebin()
        else:
            print('Load transmission spectrum, ')
            print('Then run EbinHapi.spectrum_by_ebin() to calculate previous spectra.')
	
    def save_to_tikhonov(self, suffix=''):
        '''
        Push relevant E"-binning terms to text file for nonuniform inversion.
        Send T_DB and normalized linestrength curve to MATLAB for Tikhonov regularization.
        '''
        self.ebin_output = np.transpose(np.asfarray([self.ebin_xVals, 
                                                     self.snorm_out, 
                                                     self.snorm_uc]))
        # set up the string-formatted array
        np.savetxt('temp.txt', self.ebin_output)
        with open('temp.txt','r') as out:
            ebin_str = out.readlines()
        with open(self.full_output_name + suffix + '.txt','w') as out:
            out.write('T_DB (Celsius) = %d\n' % self.t_celsius)
            out.write('lower_state_energy normalized_linestrength uncertainty\n')
            out.writelines(ebin_str)
        os.remove('temp.txt')
        
    def __cleanup__(self):
        '''
        Remove linelist files created just for this object.
        '''
#        os.remove(self.path)
#        if self.n_lines > 0:
#            os.remove(self.output_name + 'lineList.dat')
#            os.remove(self.output_name + 'lineList.dir')
#            os.remove(self.output_name + 'lineList.bak')
#            self.n_lines = 0
        try:
            [os.remove(os.path.join(self.path,f)) for f in os.listdir(self.path) 
                                       if self.output_name+'_' in f and '.data' in f]
            [os.remove(os.path.join(self.path,f)) for f in os.listdir(self.path) 
                               if self.output_name+'_' in f and '.header' in f]
        except:
            pass

###############################################################################
# Time-domain fit functions used in fit_snorm_* functions inside Ebin_Hapi class
###############################################################################
def spectra_float_width(vars, xx, molec_id, iso, temperature_kelvin, 
                             molefraction, pathlength, data, weight, 
                             shift, databases, sdvoigt):
    absorption_all = np.zeros(len(xx))
    n_bins = len(databases)
    for i, db_name in enumerate(databases):
        p_atm = vars[n_bins + i]
        alpha_ebin = calc_spectrum_fd(xx, molec_id, iso, molefraction,
                                      p_atm, temperature_kelvin,
                                      pathlength, shift[i], db_name, sdvoigt)
        absorption_all += vars[i] * alpha_ebin
    absorp_td = np.fft.irfft(absorption_all)
    return weight * (data - absorp_td)

def spectra_float_width_shift(vars, xx, molec_id, iso, temperature_kelvin,
                                   molefraction, pathlength, data, weight, 
                                   databases, sdvoigt, fix_bins, p_fix, shift_fix):
    absorption_all = np.zeros(len(xx))
    n_bins = len(databases)
    # Lineshape fit arrays
    float_bins = np.delete(np.arange(n_bins), fix_bins)
    p_atm = np.zeros(n_bins)
    p_atm[float_bins] = vars[n_bins:2*n_bins-len(fix_bins)]
    p_atm[fix_bins] = p_fix
    shift = np.zeros(n_bins)
    shift[float_bins] = vars[2*n_bins-len(fix_bins):]
    shift[fix_bins] = shift_fix
    # Simulate spectral signatures for each E"-bin
    for i, db_name in enumerate(databases):
        alpha_ebin = calc_spectrum_fd(xx, molec_id, iso, molefraction,
                                      p_atm[i], temperature_kelvin,
                                      pathlength, shift[i], db_name, sdvoigt)
        absorption_all += vars[i] * alpha_ebin
    absorp_td = np.fft.irfft(absorption_all)
    return weight * (data - absorp_td)

# nested functions for fitting
def shift_scale(y0_fd, shift, dx):
    '''
    Estimate spectral shift by interpolating from nominal cross-section.
    
    INPUTS:
        y0_fd = original frequency-domain cross-section
        shift_by_dx = desired frequency shift divided by frequency resolution
                      of original cross-section
    '''
    # avoid changing the edge-points of the spectrum
    rem = (shift % dx) / dx # is this wrong with certain shift sign?
    offset = int(shift // dx)
    # calculate xi and hwo many points to remove based on offset
    n = len(y0_fd)
    y_fd = y0_fd.copy()
    if offset < 0:
        xi = np.arange(n + offset - 1)
        y_fd[xi - offset] = y0_fd[xi] * (1 - rem) + y0_fd[xi + 1] * rem
    else:
        xi = np.arange(offset, n-1)
        y_fd[xi - offset] = y0_fd[xi] * (1 - rem) + y0_fd[xi + 1] * rem
    
    return y_fd

def pressure_scale(y0_td, gamma):
    '''
    Scale Lorentz-width of absorption cross-section.
    Using simplified method of Makowiecki et al JQSRT 2020
    '''
    # assume symmetric cepstrum y0_td
    # only scale ~ 50% of the cepstrum
    # assume the rest is below noise floor and should stay that way
    nt = len(y0_td) // 4
    y0_td[:nt] *= np.exp(gamma * np.arange(nt))
    y0_td[:-nt-1:-1] = np.exp(gamma * np.arange(nt))
    
    return y0_td

'''
Hapi support file
'''
def get_linelist_header(linelist_file):
    header_file = linelist_file.split('.')[0] + '.header'
    if os.path.exists(header_file):
        with open(header_file, 'r') as h:
            header = eval(h.read()) # load file into dictionary
        if 'extra' in header:
            print("Using non-standard HITRAN parameters for:")
            print(header['extra'])
            
        return header
    else:
        print("Assuming standard par file without .header file or extra parameters")
        
        return HITRAN_DEFAULT_HEADER

#######
# Fin #
#######



# -*- coding: utf-8 -*-
"""
Helpful functions for time-domain fitting preparation.

td = time-domain

Created on Fri Nov  8 16:46:56 2019

@author: Nate the Average
"""

import numpy as np

def largest_prime_factor(n):
    i = 2
    while i * i <= n:
        if n % i:
            i += 1
        else:
            n //= i
    return n


def bandwidth_select_td(x_array, band_fit, max_prime_factor = 500):
    '''
    Tweak bandwidth selection for swift time-domain fitting.
    
    Time-domain fit does inverse FFT for each nonlinear least-squares iteration,
    and speed of FFT goes with maximum prime factor.
    
    INPUTS:
        x_array = x-axis for measurement transmission spectrum
        band_fit = [start_frequency, stop_frequency]
    '''
    x_start = np.argmin(np.abs(x_array - band_fit[0]))
    x_stop = np.argmin(np.abs(x_array - band_fit[1]))
       
    prime_factor = largest_prime_factor(2 * (x_stop - x_start - 1))
    while prime_factor > max_prime_factor:
        x_stop -= 1
        prime_factor = largest_prime_factor(2 * (x_stop - x_start - 1))
    return x_start, x_stop

def weight_func(len_fd, bl, etalons = []):
    '''
    Time-domain weighting function, set to 0 over selected baseline, etalon range
    INPUTS:
        len_fd = length of frequency-domain spectrum
        bl = number of points at beginning to attribute to baseline
        etalons = list of [start_point, stop_point] time-domain points for etalon spikes
    '''
    weight = np.ones(2*(len_fd-1))
    weight[:bl] = 0; weight[-bl:] = 0
    for et in etalons:
        weight[et[0]:et[1]] = 0
        weight[-et[1]:-et[0]] = 0
    return weight


# -*- coding: utf-8 -*-
import numpy as np
import os # or pathlib
from time import strftime, localtime
try:
    from .td_support import bandwidth_select_td
except:
    from td_support import bandwidth_select_td
    
class Transmission():
    def __init__(self, bandwidth=[6800,7100]):
        
        # transmission spectrum information
        self.trans_spectrum = None
        self.x_wvn = None
        self.wvn_start, self.wvn_stop = bandwidth
        self.transmission_path = '' # what file did transmission spectrum come from?
        
        # initialize thermodynamic parameters
        self.p_torr = 640
        self.t_celsius = 23
        self.chi = 0.01
        self.pathlength = 10 
                
    def get_transmission_asc(self, asc_file_path, overwrite_thermo=True):
        """
        Pulls important parameters from transmission file already formatted for Labfit
        
        INPUTS:
            asc_file_path = string name of transmission file (ignore .asc extension)
                            must be in labfit_engine directory else Labfit won't run
            self.wvn_start = fitting range for parameter
            
        OUTPUTS:
            Numpy arrays of frequency-domain transmission spectrum in object
            self.x_wvn = frequency axis (wavenumbers cm-1)
            self.trans_spectrum = intensity transmission (frequency-domain)
            
            Also pulls path-average thermo properties T,P,X,L from file
            self.pathlength
            self.t_celsius
            self.p_torr
            self.chi        
        """
        
        if '.asc' in asc_file_path:
        # Ignore .asc extension
            asc_file_path = asc_file_path[:-4]        
        self.transmission_path = os.path.abspath(asc_file_path) + '.asc'
        
        ## First extract path-average thermodynamic parameters
        if overwrite_thermo:
            with open(asc_file_path + '.asc','r') as f:
                for i in range(3):
                    f.readline()
                therm_line = f.readline()
                path_m, t_celsius, p_torr, chi = therm_line.split()[:4]
                self.pathlength = float(path_m) * 100 # cm
                self.t_celsius = float(t_celsius)
                self.p_torr = float(p_torr)
                self.chi = float(chi)

        ## Then get transmission spectrum
        self.x_wvn, self.trans_spectrum = np.loadtxt(asc_file_path + '.asc',unpack=True, skiprows=18)
        # last update the subset bandwidth
        # Check that transmission spectrum is large enough
        if self.x_wvn[0] > self.wvn_start:
            self.wvn_start = np.ceil(self.x_wvn[0])
            print('Bandwidth too large for transmission file.')
            print('Changing self.wvn_start = ', self.wvn_start)
        if self.x_wvn[-1] < self.wvn_stop:
            self.wvn_stop = np.floor(float(self.x_wvn[-1]))
            print('Bandwidth too large for transmission file.')
            print('Changing self.wvn_stop = ', self.wvn_stop)
        # Adjust array to match fit window with fast FFT for time-domain 
        x_start, x_stop = bandwidth_select_td(self.x_wvn, [self.wvn_start, self.wvn_stop])
        self.x_wvn = self.x_wvn[x_start:x_stop]
        self.trans_spectrum = self.trans_spectrum[x_start:x_stop]
        
    def make_asc(self, file_name = 'Spectrum007',
                 message = 'Data for E"-bin time-domain fitting', save_res = False):
        """
        Make Labfit-formatted transmission text file.
        18 lines of header, then space-delimited array x_wvn, trans
        
        INPUTS:
            self.trans_spectrum = laser transmission spectrum (y-axis numpy array)
            self.x_wvn = (cm-1) wavenumber of each data point (numpy array)
            self.p_torr, self.t_celsius, self.pathlength, self.chi = path-average properties
            file_name = name of output ASC file for Labfit
            save_res: if True, save absorbance residual in 3rd column (from self.spectrum_by_ebin())
            
        OUTPUTS:
            transFile.asc = Labfit-formatted transmission spectrum
            
        ASSUMPTIONS:
            self.wvn_start and self.wvn_stop are where you want them
            Actually Labfit can handle an ASC file larger than the fitting region
        
        """
        
        self.transmission_path = os.path.abspath(file_name) + '.asc' # !! possible bug: remove .txt or other file extension

        ## Format strings for ASC file conventions
        self.wvn_step = (self.x_wvn[-1] - self.x_wvn[0]) / (len(self.x_wvn) - 1) # :31f
        # Change T,P,X,L to formatted strings
        pathlength = self.pathlength/100; T_celsius = self.t_celsius; p_torr = self.p_torr
        if(pathlength < 10):
            pathlength = '{0:.7f}'.format(pathlength)
        elif(pathlength < 100):
            pathlength = '{0:.6f}'.format(pathlength)
        if(T_celsius < 100): # Assume T_celsius > 10c
            T_celsius = '{0:.4f}'.format(T_celsius)
        elif(T_celsius < 1000):
            T_celsius = '{0:.3f}'.format(T_celsius)
        else:
            T_celsius = '{0:.2f}'.format(T_celsius)
        if(p_torr < 10):
            p_torr = '{0:.5f}'.format(p_torr)
        elif(p_torr < 100):
            p_torr = '{0:.4f}'.format(p_torr)
        elif(p_torr < 1000):
            p_torr = '{0:.3f}'.format(p_torr)
        elif(p_torr < 10000):
           p_torr = '{0:.2f}'.format(p_torr)
        else:
            p_torr = '{0:.1f}'.format(p_torr)
        chi = '{0:.6f}'.format(self.chi)
        
        ## Now write to output file
        with open(self.transmission_path,'w') as out_file:
            file_id = '1111'
            out_file.write('******* File ' + file_id + ', ' + message + '\n')
            out_file.write(file_id+'   ')
            out_file.write(f'{self.x_wvn[0]:.10f}' + '  ')
            out_file.write(f'{self.x_wvn[-1]:.8f}'  + '  ')
            out_file.write( f'{self.wvn_step:.31f}' + '\n')
            out_file.write('  00000.00    0.00000     0.00e0     ')
            out_file.write('1   2     3   0        0\n') # Hitran molecule codes
            # Thermo line
            out_file.write(4*' ' + pathlength + 5*' ' + T_celsius + 6*' ' + 
                           p_torr + 4*' ' + chi + '    .0000000 .0000000 0.000\n')
            for i in range(2):
                out_file.write('    0.0000000     23.4486      0.00000    0.000000')
                out_file.write('    .0000000 .0000000 0.000\n')
            for i in range(9):
                out_file.write('1. 1. 1. 1. 1. 1. 1. 1. 1. 1.\n')
        # Other introductory lines
#            out_file.write('DATE ' + date + '; time ' + time + '\n')
            out_file.write(strftime("%m/%d/%Y %H:%M:%S\n", localtime()))
            out_file.write('\n')
            out_file.write('6801.3716254 0.003338200000000000000000000000 15031')
            out_file.write(' 1 1    2  0.9935  0.000   0.000 294.300 295.400')
            out_file.write(' 295.300   7.000  22.000 500.000 START\n')
         # Now write out transmission spectrum
            for count, vals in enumerate(self.x_wvn):
                out_file.write(f'{vals:.5f}' + 6*' ')
                out_file.write(f'{self.trans_spectrum[count]:.5f}')
                if save_res:
                    res_i = self.data_meas[count] - self.absorption_total[count]
                    out_file.write(6*' ' + f'{res_i:.5f}')
                out_file.write('\n')



# -*- coding: utf-8 -*-
"""
Fit a transmission spectrum from some ASC file for temperature nonuniformity.

Produces a discrete normalized linestrength curve with statistical fitting uncertainties.
which can be then be imported into Tikhonov regularization algorithm (Ref 1)
 to solve for the nonuniform temperature distribution.
This E"-binning algorithm (Ref 2) uses time-domain fitting (Ref 3) 
 to extract concentrations at different E"-ranges.

PREREQUISITE:
This fitting code produces model spectra using the external hapi package (Ref 4).
This package includes a modification of hapi, hapi_combustion.py
to produce more accurate high-temperature lineshape models.

This package uses object-oriented programming. 
You can tab-complete in IPython console to see which functions are available.
INPUTS:
    ( see 5-step preparation in script below to see how to implement )
    - transmission spectrum (and corresponding x-axis)
    - absorption linelist (H2O works best)
    - path-average TPXL to calculate absorption lineshapes
    - which bins of lower-state energy to fit as separate concentrations
OUTPUTS:
	- Normalized linestrength curve text file, in this instance named "example_td.txt"
	   For input into TikhonovInversion algorithm (Ref 1)
       Information also stored in object as N.snorm_out and N.t_celsius
	- Plot of scaled concentration fits of all the E"-bins.


References:
1: Malarich, Rieker, "Part I: ..."
2: Malarich, Rieker, "Part II: ..."
3: Makowiecki, Cole, Hoghooghi, Rieker. ""
4: R.V. Kochanov, I.E. Gordon, L.S. Rothman, P. Wcislo, C. Hill, J.S. Wilzewski, 
   HITRAN Application Programming Interface (HAPI): A comprehensive approach to working 
   with spectroscopic data, J. Quant. Spectrosc. Radiat. Transfer 177, 15-30 (2016)
Created on Tue Jul 23 15:15:45 2019

@author: ForAmericanEyesOnly
"""
# built-in modules
import numpy as np
import matplotlib.pyplot as plt
import os

# nonuniform-temperature code
#from sys import path as syspath
#syspath.append(r'../')
from ntfit import spectrafit as ebin
from ntfit.tdist import Snorm2Tx
plt.rcParams.update({'figure.autolayout':True,'lines.linewidth':0.8}) # plot formatting

load_previous = False
new_fit = True
temperature_inversion = False

# Produce file + path name for computation and output files
file_name = os.path.join(os.getcwd(),'data','example_td')

transmission_labfit = r'data\Furnace_Synth.asc'# ASC file with transmission and thermodynamic data
# Other parameters to edit as you choose.
band_fit = [6802, 7000] # fit region (cm-1)
recommended_iteration = False # Get more accurate higher linestrengths for E"<1500 by floating each broadening. 
calc_nonuni_spectrum = True # Calculate transmission spectrum used in this example fit.

if load_previous:
    # previous E"-bin fit from transmission spectrum (in .asc file)
    NOld = ebin.EbinHapi(file_name, bandwidth = band_fit)
    NOld.load_fit(load_transmission=True, load_path_override = r'data\example_td_ebin')

    # previous T(x) fit from S(E")
    LOld = Snorm2Tx()
    LOld.load_previous_fit(r'data\example_td_tx')
    LOld.plot_fit();


if calc_nonuni_spectrum:    
    N = ebin.EbinHapi(r'data/furnace_synth', bandwidth = band_fit)
    # 2) Select linelist to fit
    N.linelist_path = r'linelist/H2O_PaulLF.data' # Hitran-formatted PAR file
                         # must also have .header file of same name to run HAPI
    # Produce synthetic spectrum from known T(x) profile
    # How to calculate the transmission spectrum used in Furnace_Synth.asc
    N.tx_act_kelvin = np.asarray([547, 581, 634, 699, 771, 843, 908, 959, 991, 996, 
                                969, 903, 877, 822, 772, 724, 680, 638, 600, 564, 
                                531, 500, 471, 443, 418, 395, 372, 351, 331, 312])
    N.pathlength = 152.4 #cm
    N.t_celsius = np.round(np.average(N.tx_act_kelvin) - 273.15)
    N.x_wvn = np.arange(6802, 7000,.01) # 
    abs_nonuni, abs_uni = N.simulate_spectra(True)
    plt.figure(); plt.plot(N.x_wvn, abs_nonuni, label = 'nonuniform spectra')
    plt.plot(N.x_wvn, abs_uni, label = 'uniform-spectra approximation')
    plt.xlabel('Frequency (cm-1)'); plt.ylabel('Absorbance'); plt.legend()
    # data[:,1] = 100 * np.exp(-abs_nonuni) # how Furnace_Synth.asc transmission calculated
    
#    # Now load simulated spectrum into E"-bin fit
#    N.x_wvn = np.arange(6802, 7000,.01)
#    N.trans_spectrum = 100 * np.exp(-abs_nonuni)
#    N.print_thermo() # check that TPXL consistent with previous fit
    
    N.__cleanup__()


if new_fit:
    ##############################################################################
    ## Object-oriented time-domain fitting. 3-5 steps of preparation before fitting.
    # 1) Initialize the E"-binning object. 
    N = ebin.EbinHapi(file_name + 'new', bandwidth = band_fit)
    
    # 2) Select linelist to fit
    N.linelist_path = r'linelist/H2O_PaulLF.data' # Hitran-formatted PAR file
                         # must also have .header file of same name to run HAPI
    
    # 3) Pull in transmission spectrum and estimated path-average thermo parameters
    if calc_nonuni_spectrum:
        # standard method, if you don't already have ASC file
        N.x_wvn = np.arange(6802, 7000,.01)
        N.trans_spectrum = 100 * np.exp(-abs_nonuni)  
        # and update the path-average TPXL fits
        N.t_celsius = 380
        N.pathlength = 152.4
    else:
    # Furnace_Synth.asc file contains transmission spectrum in (wavenumber (cm-1), transmission),
    #  as well as estimated path-average thermodynamic parameters:
    #       p_torr, t_celsius, chi, pathlength (cm)
        N.get_transmission_asc(r'data\Furnace_Synth.asc')
    # 4) This function also updates thermodynamic parameters from asc file
    
    # 5) (optional) Update E"-bin selections.
    # The linelist is split into several different "molecules",
    # so each "molecule" has transition lower-state energies within these delineations
    N.Elist = [0,200,500,750,1020,1220,1440,1680,1940,2220,2510,2820,3100,7000]
    
    # 6) (optional) Update baseline and etalon list
    # N.bl = 100; N.etalons = [[100,120],[150,180]]
    
    ##############################################################################
    ## Now the actual fitting and analysis
    # Select any fit_snorm_* function for different amount of fit parameters
    # Fit concentrations of each E"-bin
    N.fit_snorm(); # the fastest fit, only float pathlength of each E"-bin.
    # All fit_ functions run N.spectrum_by_ebin() to make figure plot.
    #  and calculate baseline-removed spectrum N.data_meas
    # If plot is unseemly and slow, can also do N.plot_ebin() after running a fit_*()
    foo = N.fit_temperature() # plot linestrength curve and show nonuniformity evidence
 
    if load_previous:
        match_previous = np.allclose(NOld.snorm_out, N.snorm_out, .02) 
        if match_previous:
            print('Fit matches previous')
        else:
            print('New fit inconsistent with previous file')
    
## Optional re-fit of concentrations, also floating pressure-broadening of each E"-bin.
# This step is often critical to reduce linestrength error for E"<1500.
#  (see Ref 2, section 2.3: reducing lineshape error)
    if recommended_iteration:
        foo = N.fit_snorm_width_shift()
        print('Normalized linestrength (area/width fit)',N.snorm_out)
        print('Fitted P=',N.p_torr)
        
## Keep clean directory
# Save your fit to reload in another session
N.save_fit()
# Remove temporary linelist files from directory.
N.__cleanup__() # Must repeat N.get_linelist_par() to make another fit.

if temperature_inversion:
    # Now get temperature distribution from linestrengths
    L = Snorm2Tx(N)
    L.reg_weight = 10.**(np.arange(2,-5,-1))
    plot_handles = L.length_bin()
    L.save_fit()


# -*- coding: utf-8 -*-
"""
Simulated annealing algorithm for Tikhonov regularization.
Get discretized temperature distribution from set of normalized linestrengths.

To use, write
tx_, res_, pxl_, reg_, corner = lbin_sa(meas, model)

Or object-oriented version:
L = Snorm2Tx(Ebin_Object)
plot_handles = L.length_bin()
L.save_fit()

MATH abbreviations:
sa = simulated annealing
reg = Tikhonov regularization
l2 = L2-norm matrix for 2nd-order Tikhonov regularization, ||d2f / dx2||
lsq = least-squares
n_ = number of thing
fx = f(x) objective function to minimize
uc = uncertainty

SPECTROSCOPY abbreviations
e, elower = lower-state energy E" (cm-1)
snorm = normalized linestrength
t = temperature (Kelvin)
db = database (aka reference). 
     t_db is reference temperature for linestrength-fitting
tx = T(x) temperature vs position profile, a length-bin discretization
pdf = Probability Density Function of pxl at each temperature
lbin = length-bin, where temperature profile split into several equal lengths.
uni = uniform-path, single-temperature fit
pxl = pressure*molefraction*pathlength product, (cm * atm)
      some count of number absorbing molecules

Development notes:
The central fitting code, tikhonov_sa comes from Corana et al[1].
My implementation for temperature nonuniformity is based on Cai et al [2]
I haven't seen that code, but the 2 ways this might be different is stability
for a single-path absorption measurement (rather than tomography),
with (1) the w_ub term in simulated annealing to encourage parabolic profiles,
and also (2) the L-curve as a lever between uniform and least-square profiles,
giving some uncertainty estimates in profile. These details in [3].

References:
[1] A Corana, M Marchesi, C Martini, S Ridella, "Minimizing multimodal functions
    of continuous variables with the 'simulated annealing' algorithm," 
ACM Trans. Math. Softw. 13 (3) 1987
[2] W Cai, D Ewing, L Ma, "Application of simulated annealing for multispectral
    tomography," Comp. Phys. Comm. 179 (4) 2008
[3] NA Malarich, GB Rieker, "Resolving nonuniform temperature distributions with
 single-beam absorption spectroscopy I) Theoretical capabilities and limitations,"
    JQSRT
    
Created on Sun Dec 22 10:41:41 2019

@author: Nathan Malarich
"""

import os
import numpy as np
from numpy.linalg import norm
#from numpy.matlib import repmat
from time import time
import matplotlib.pyplot as plt
import json
#from multiprocessing import Pool
from scipy.optimize import nnls

# partition function TIPS2017 from hapi (available HitranOnline)
from .hapi import partitionSum
FOUND_HAPI = True
HC_K = 1.4388028496642257

class Snorm2Tx():
    
    def __init__(self, EbinHapi=None):
        '''
        INPUT:
            EbinHapi spectral fitting object
            will pull all necessary variables from spectral fit object,
            and save to same directory with same basename.
            
            If you don't have an object handy in your IPython console, you can
            1) populate the 5 values manually
            2) Load a previous spectral fit from your hard drive
               >> N = ebin.EbinHapi(full_save_path)
               >> N.load_fit()
               >> L = Snorm2Tx(N)
        '''
        # E"-binning fit results
        if EbinHapi is not None:
            self.t_db = EbinHapi.t_celsius + 273
            self.elower = EbinHapi.ebin_xVals
            self.snorm = EbinHapi.snorm_out
            self.snorm_uc = EbinHapi.snorm_uc
            self.save_name = EbinHapi.full_output_name
            # and variables for Q(T)
            self.molec_id = EbinHapi.molec_id
            self.local_iso_id = EbinHapi.local_iso_id
        else:
            self.t_db = 0
            self.elower = None
            self.snorm = None
            self.snorm_uc = None
            self.save_name = None
            self.molec_id = 1
            self.local_iso_id = 1            
            print('Calculating linestrengths with H2O partition function')
            print('Use load_ebin() or load_previous_fit() before running other functions')
        
        self.model = None
        
        # length-binning parameters
        self.nbins = 16
        self.start_condition = 'uniform'
        self.tmin = 300
        self.tmax = max(1300, self.t_db + 400)
        self.reg_weight = 10.**-np.arange(5)
        self.corner_thrsh = 0.07
        
        # extreme-fit outputs
        self.tx_uni = None
        self.tx_lsq = None
        
        # length-binning outputs
        self.tx = None
        self.res = None
        self.pxl = None
        self.reg = None
        self.corner = None
    
    def length_bin(self, tmax_override = None):
        '''
        Run length-bin simulated annealing code.
        
        Same algorithm as lbin_sa(), 
        except outputs stay in Snorm_2Tx object instance
        '''
        ## For minimize ||Ax - b||, calculate matrix A and solution b
        # calculate weighted inversion matrix self.model_w
        model = self.get_model(tmax_override)
        meas_w = self.snorm / self.snorm_uc # solution vector
        
        # Calculate extreme-condition inversions
        self.tx_uni, self.res_uni, self.pxl_uni = uniform_fit(meas_w, self.model_w,
                                                              self.tmin, self.nbins)
        self.tx_lsq, self.res_lsq, self.pxl_lsq = lsq_fit(meas_w, self.model_w,
                                                          self.tmin, self.nbins)
        l2 = self.get_l2()
        self.reg_lsq = norm(l2 @ self.tx_lsq)
        if self.start_condition == 'uniform':
            tx0 = self.tx_uni
            pxl0 = self.pxl_uni
        else:
            tx0 = self.tx_lsq
            pxl0 = self.pxl_lsq
        
        # Now calculate Tikhonov inversion at several regularization weights
        self.tx = np.zeros((self.nbins, len(self.reg_weight)))
        self.pxl = np.zeros(len(self.reg_weight))
        self.res = np.zeros(len(self.reg_weight))
        self.reg = np.zeros(len(self.reg_weight))
        self.s_fit = np.zeros((len(self.elower), len(self.reg_weight)))
        tic = time()
        for i, reg in enumerate(self.reg_weight):
            self.tx[:,i], self.pxl[i], self.res[i] = tikhonov_sa(tx0, pxl0, meas_w, 
                   self.model_w, reg, nbins = self.nbins, tmin = self.tmin)
            self.s_fit[:,i] = res(self.tx[:,i], self.pxl[i], model)
            self.reg[i] = norm(l2 @ self.tx[:,i])
        print('Tikhonov regularization in %d seconds' % (time() - tic))
        fig, axs = self.plot_fit()
        
        return fig, axs
    
    '''
    Nest functions called within length_bin()
    '''    
    def plot_fit(self, plot_limits = False, vstack = False):
        '''
        Plots output of self.length_bin() and self.lcurve()
        INPUTS (optional):
            plot_limits = plot uniform and least-squares solutions as well
            vstack = stack subplots vertically (default horizontal)
        '''
        # edge cases
        if self.model is None:
            self.get_model()
        self.s_uni = res(self.tx_uni, self.pxl_uni, self.model)
        self.s_lsq = res(self.tx_lsq, self.pxl_lsq, self.model)
        
        bad_fits = self.lcurve()
        if vstack:
            fig, axs = plt.subplots(3,1)
            title_plot = 0
        else:
            fig, axs = plt.subplots(1,3)
            title_plot = 1
        # plot linestrengths
#        axs[0].set_title('Spectroscopy data')
        if plot_limits:
            axs[0].plot(self.elower, self.s_uni, 'k--', label='uniform')
            axs[0].plot(self.elower, self.s_lsq, 'k:', label='least-squares')
        if len(bad_fits) > 0:
            axs[0].plot(self.elower, self.s_fit[:, bad_fits], linewidth=0.5)
        axs[0].plot(self.elower, self.s_fit[:, self.corner_all], linewidth=2)
        axs[0].errorbar(self.elower, self.snorm, self.snorm_uc, color='k', ls='none')
        axs[0].plot(self.elower, self.snorm, 'k.', label = 'Data')
        axs[0].set_xlabel('Lower-state energy E"')
        axs[0].set_ylabel('Normalized linestrength')    
        
        # plot temperature distributions
#        axs[1].set_title('Temperature distribution')
        x = np.linspace(0,1, self.nbins)
        if plot_limits:
            axs[1].plot(x, self.tx_uni,'k--')
            axs[1].plot(x, np.sort(self.tx_lsq),'k:')
        for b in bad_fits:
            axs[1].plot(x, np.sort(self.tx[:,b]), linewidth=0.5)
        for c in self.corner_all:
            axs[1].plot(x, np.sort(self.tx[:, c]), linewidth=2)
        axs[1].set_xlabel('Column density fraction')
        axs[1].set_ylabel('Temperature (K)')
        axs[1].set_ylim(self.tmin, self.tmax)
        axs[1].set_xlim(0, 1)
        try:
            axs[title_plot].set_title(os.path.basename(self.save_name))
        except:
            pass
        
        # plot L-curve
#        axs[2].set_title('Tikhonov L-curve')
        for b in bad_fits:
            axs[2].plot(self.res[b], self.reg[b], '+')
        for c in self.corner_all:
            axs[2].plot(self.res[c], self.reg[c], '*')
        axs[2].set_xlabel('Linestrength residual')
        axs[2].set_ylabel('Temperature curvature')
        # Add uniform and leastsquare bounds to L-curve
        axs[2].plot(self.res_uni, 0, 'kx')
        axs[2].plot(min(self.res_lsq, min(self.res)), self.reg_lsq, 'kx')
        
        return fig, axs
                   
    def lcurve(self, thrsh_change = None):
        '''
        Calculates L-curve and decides which T(x) solutions within uncertainty.
        
        See Appendix 1 of Ref [3]        
        '''
        # Set up L-curve rectangle equation (Eq. A1 in ref [3])
        reg_uni = 0
        # bug fix: by discretizing tx_lsq, self.res_lsq could be higher residual
        res_lsq = min(self.res_lsq, min(self.res))
        y_reg = (self.reg - reg_uni) / self.reg_lsq
        x_res = (self.res - res_lsq) / self.res_uni
        reg_rectangle = np.sqrt( y_reg**2 + x_res**2 )
        self.corner = int(np.argmin(reg_rectangle))
        # corner_all are all solutions close enough to true L-curve corner.
        self.corner_all = []
        not_corner = []
        if thrsh_change is not None:
            self.corner_thrsh = thrsh_change
        for index, r in enumerate(reg_rectangle):
            if r < self.corner_thrsh:
                self.corner_all.append(index)
            else:
                not_corner.append(index)
        if len(self.corner_all) == 0:
            self.corner_all.append(self.corner)
        self.corner_all = np.asarray(self.corner_all).astype(int)
        not_corner = np.asarray(not_corner).astype(int)
        # Residual improvement from allowing nonuniformity
        res_corner = self.res[self.corner]
        nonuni_improvement = (self.res_uni - res_corner) / self.res_uni * 100
        print(int(nonuni_improvement), '% fit improvement from uniform-fit')
        
        return np.asarray(not_corner)

    def get_l2(self):
        '''
        Nested function to calculate matrix for 2nd-order Tikhonov regularization.
        
        Minimizes finite-difference d2T/dx2
        '''
        l2 = 2 * np.diag(np.ones(self.nbins))
        l2 -= np.diag(np.ones(self.nbins-1),1) + np.diag(np.ones(self.nbins-1),-1)
        l2[0,:] = 0; l2[-1,:] = 0 # ignore edges
        l2[-1,-1] = 2; l2[-1,-2] = -2 # add 2nd-derivative to high-temperature edge
        
        return l2
    
    def calc_hessian(self):
        '''
        Calculate nonlinear-least-squares Hessian matrix for length-bin result.
        
        H = A'A + gamma * L2'L2
        Weighted model matrix derivative A' = (1/snorm_uc_i) * d(snorm(Tj))/d(Tj)
        '''
        assert self.corner is not None, "run self.length_bin() first to get tx result."
        
        a = np.zeros((len(self.snorm), self.nbins))
        tx_best = self.tx[:,self.corner]
        for j, tj in enumerate(tx_best):
            j2 = int(min(tj - self.tmin, np.shape(self.model_w)[1] - 1))
            a[:,j] = self.model_w[:,j2+1] - self.model_w[:,j2]
        ata = np.transpose(a) @ a
        l = self.get_l2()
        ltl = np.transpose(l) @ l
        gamma = self.reg_weight[self.corner] # No rescaling
        
        return ata + gamma**2 * ltl
        # uc = np.sqrt(np.diag(np.linalg.inv(hessian))) ? 
        
    
    def get_model(self, tmax_override=None):
        '''
        Nested function to get normalized linestrength matrix
        INPUTS:
            tmax_override = maximum temperature-column of matrix (int/float)
            q_exact = use exact partition function from TIPS 
                       instead of power-law (Boolean)
        
        model_ij = integrated_area(E"_i) / (S_i(T_j) PXL)
        '''
        help_statement = "Load linestrengths with some load_*() function"
        help_statement += "before calling this function again."
        assert self.t_db != 0, help_statement
        # Set up absorption model
        
        # and model matrix
        if tmax_override is not None:
            self.tmax = tmax_override
        t = np.arange(self.tmin, self.tmax + 1)
        self.model_w = np.zeros((len(self.elower), len(t)))
        self.model = self.model_w.copy()
        for index, e in enumerate(self.elower):
            self.model[index,:] = s_t_exact(t, self.t_db, e, 
                                           self.molec_id, self.local_iso_id)
            self.model_w[index,:] = self.model[index,:] / self.snorm_uc[index]
#        snorm_repmat = np.transpose(repmat(self.snorm_uc, len(t),1))
#        model = self.model_w * snorm_repmat       
            
        return self.model
    
    def calc_uc(self, reg_index = None):
        '''
        Estimate statistical fitting uncertainty from Tikhonov result
        INPUT:
            reg_index = (int) which column of self.tx to calculate uncertainty.
                        default self.corner
                        
        TODO:
            Figure out why singular matrix for uniform retrieval.
            Is this invalidation of the Hessian derivation or method?
        '''
        assert self.tx is not None, "run self.length_bin() first to get tx result."
        
        if reg_index is None:
            reg_index = self.corner
        
        a = np.zeros((len(self.snorm), self.nbins + 1))
        tx_best = self.tx[:,reg_index]
        for j, tj in enumerate(tx_best):
            j2 = int(min(tj - self.tmin, np.shape(self.model_w)[1] - 2))
            a[:,j] = self.model_w[:,j2+1] - self.model_w[:,j2]
            a[:,-1] += self.model_w[:,j2] # d(snorm) / d(pxl)
        ata = np.transpose(a) @ a
        
        l = self.get_l2()
        l2 = np.zeros((self.nbins, self.nbins + 1)) # with extra pxl column
        l2[:,:-1] = l
        ltl = np.transpose(l2) @ l2
        gamma = self.reg_weight[reg_index] # No rescaling
        
        hessian = ata + gamma**2 * ltl
        uc = np.sqrt(np.diag(np.linalg.inv(hessian)))
        
#        # Uniform retrieval case, ignore possibility of nonuniformity
#        if max(tx_best) == min(tx_best):
#            a = np.zeros((len(self.snorm),2))
#            tj = tx_best[0]
#            j2 = int(min(tj - self.tmin, np.shape(self.model_w)[1] - 2))
#            a[:,0] = self.model_w[:,j2+1] - self.model_w[:,j2] # d(snorm) / dT
#            a[:,1] = self.model_w[:,j2] # d(snorm) / d(pxl)
#            ata = np.transpose(a) @ a
#            uc = np.sqrt(np.diag(np.linalg.inv(ata)))

        return uc
    
    def plot_uc(self, accepted_only = False):
        '''
        Plot temperature distribution (CDF) with statistical uncertainty.
        '''
        colors ='bgrcmyk'
        plt.figure('CDF_uc')
        if accepted_only:
            cdf_min = np.zeros((self.nbins, len(self.corner_all)))
            cdf_max = np.zeros((self.nbins, len(self.corner_all)))
            for i, r in enumerate(self.corner_all):
                tx = self.tx[:,r]
                uc = self.calc_uc(r)
                cdf = np.sort(tx)
                cdf_uc = uc[cdf.argsort()]
                plt.plot(cdf,'-',color=colors[r])
                plt.plot(cdf + cdf_uc,'--',color=colors[r])
                plt.plot(cdf - cdf_uc,'k--',color=colors[r])
                cdf_min[:,i] = cdf - cdf_uc
                cdf_max[:,i] = cdf + cdf_uc
            # Take total uncertainty to be span of each length-bin retrieval
            cdf_min = np.min(cdf_min, axis=1)
            cdf_max = np.max(cdf_max, axis=1)
            return cdf_min, cdf_max
        else:
            for r in range(len(self.reg_weight)):
                tx = self.tx[:,r]
                uc = self.calc_uc(r)
                cdf = np.sort(tx)
                cdf_uc = uc[cdf.argsort()]
                plt.plot(cdf,'-',color=colors[r])
                plt.plot(cdf + cdf_uc,'--',color=colors[r])
                plt.plot(cdf - cdf_uc,'k--',color=colors[r])
    
    '''
    Data-handling functions
    '''
    def save_fit(self, suffix = ''):
        '''
        Save fit result as dictionary to text file. 
        Can reload this information using load_previous_fit()
        
        INPUTS:
            suffix = addition to text file name, 
                     to indicate multiple fits to same data.
        '''
        # setup parameters
        fit = {'t_db':self.t_db}
        fit['elower'] = self.elower.tolist()
        fit['snorm'] = self.snorm.tolist()
        fit['snorm_uc'] = self.snorm_uc.tolist()
        fit['molec_id'] = self.molec_id
        fit['local_iso_id'] = self.local_iso_id
        fit['tmin'] = self.tmin
        fit['tmax'] = self.tmax
        if type(self.reg_weight) is not list:
            fit['reg_weight'] = self.reg_weight.tolist()
        else:
            fit['reg_weight'] = self.reg_weight
        # Tikhonov inversions
        fit['tx'] = np.transpose(self.tx).tolist()
        fit['pxl'] = self.pxl.tolist()
        fit['res'] = self.res.tolist()
        fit['reg'] = self.reg.tolist()
        fit['corner'] = int(self.corner)
        # edge cases
        fit['tx_uni'] = self.tx_uni.tolist()
        fit['pxl_uni'] = self.pxl_uni
        fit['res_uni'] = self.res_uni
        fit['tx_lsq'] = self.tx_lsq.tolist()
        fit['pxl_lsq'] = self.pxl_lsq
        fit['res_lsq'] = self.res_lsq
        fit['reg_lsq'] = self.reg_lsq
        with open(self.save_name + '_tx' + suffix,'w') as f:
            f.write(json.dumps(fit, indent=2))
            
        # Also save accepted length-bin fits to separate text file
        np.savetxt(self.save_name + '_tx' + suffix + '.txt', self.tx[:,self.corner_all])
        
    def load_previous_fit(self, file_name):
        '''
        Load length-binning fit from text file (from previous Snorm2Tx.save_fit())
        
        Can then run Snorm2Tx.lcurve() or Snorm2Tx.plot_fit()
        INPUTS:
            file_name = string name of file produced with earlier length-bin fit.
        '''
        with open(file_name,'r') as f:
            fit = json.loads(f.read())
        self.save_name = file_name
        self.t_db = fit['t_db']
        self.elower = np.asarray(fit['elower'])
        self.snorm = np.asarray(fit['snorm'])
        self.snorm_uc = np.asarray(fit['snorm_uc'])
        self.tmin = fit['tmin']
        self.tmax = fit['tmax']
        self.reg_weight = np.asarray(fit['reg_weight'])
        # Tikhonov inversions
        self.tx = np.transpose(np.asarray(fit['tx']))
        self.nbins = np.shape(self.tx)[0]
        self.pxl = np.asarray(fit['pxl'])
        self.res = np.asarray(fit['res'])
        self.reg = np.asarray(fit['reg'])
        self.corner = fit['corner']
        # edge cases
        self.tx_uni = np.asarray(fit['tx_uni'])
        self.pxl_uni = fit['pxl_uni']
        self.res_uni = fit['res_uni']
        self.tx_lsq = np.asarray(fit['tx_lsq'])
        self.pxl_lsq = fit['pxl_lsq']
        self.res_lsq = fit['res_lsq']
        self.reg_lsq = fit['reg_lsq']
        # Partition function
        try:
            self.molec_id = fit['molec_id']
            self.local_iso_id = fit['local_iso_id']
        except:
            self.molec_id = 1
            self.local_iso_id = 1
        
        # and calculate previous fit residuals
        model = self.get_model()
        self.s_fit = np.zeros((len(self.elower), len(self.reg_weight)))
        for i, reg in enumerate(self.reg_weight):
            self.s_fit[:,i] = res(self.tx[:,i], self.pxl[i], model)  
     
    def load_ebin(self, Ebin):
        '''
        Load EbinHapi object to later fit with self.length_bin()
        In case you forgot when you initialized this object.
        '''
        try:
            self.t_db = Ebin.t_celsius + 273
        except AttributeError:
            self.t_db = Ebin.temp_c + 273
        self.elower = Ebin.ebin_xVals
        self.snorm = Ebin.snorm_out
        self.snorm_uc = Ebin.snorm_uc
        self.save_name = Ebin.full_output_name        


'''
Nested codes in the setup of Snorm2Tx.length_bin
'''
def s_t(t_new, t0, elower, power=-2.77):
    '''
    Returns linestrengths at s(t_new) given array E"
    
    Used in Snorm2Tx.get_model
    '''
    return (t_new / t0)**power * np.exp(-HC_K * elower * (1/t_new - 1/t0))

def s_t_exact(t_new, t0, elower, molec_id = 1, iso = 1):
    '''
    Returns linestrengths at s(t_new) given array elower and partition function q
    '''
    q_t0 = partitionSum(molec_id, iso, t0)
    q_tnew = np.asarray(partitionSum(molec_id, iso, list(t_new)))
    boltz_t0 = np.exp(-HC_K * elower / t0) / (q_t0 * t0)
    boltz_t_new = np.exp(-HC_K * elower / t_new) / (q_tnew  * t_new)
    return boltz_t_new / boltz_t0

def uniform_fit(meas, model, tmin = 300, nbins = 30):
    '''
    What is single best T,X to fit normalized linestrengths?
    '''
    n_t = np.size(model,1)
    pxl_t = np.zeros(n_t)
    res_t = np.zeros(n_t)
    for t in range(n_t):
        pxl_t[t], res_t[t] = nnls(model[:,t:t+1], meas)
    t_uni = np.argmin(res_t) + tmin
    pxl_uni = pxl_t[t_uni - tmin]
    res_uni = res_t[t_uni - tmin] # 2-norm residual
    t_uni = t_uni * np.ones(nbins)
    
    return t_uni, res_uni, pxl_uni

def lsq_fit(meas, model, tmin = 300, nbins = 30):
    '''
    Get PDF(T) least-squares solution and turn into symmetric discrete profile.
    INPUTS:
        meas = normalized linestrengths vector
        model = inversion matrix (E" vs T)
        temp_array = (Kelvin) temperatures used for columns of model
        nbins = number of spots to split temperature
    TODO:
        lower temperature resolution 5 Kelvin?
    '''
    tx = np.zeros(nbins) # length-bin
    
    pdf, res_lsq = nnls(model, meas)
    pxl_lsq = sum(pdf)
    pdf *= nbins / pxl_lsq
#    cdf = np.zeros(len(pdf))
    
    # now round the continuous-numbered PDF into length bins
    lbin = 0; # length-bin index
    res_bin = 1
    # keep a weighted average of pdf until you hit next discretization point
    for ti in range(np.size(model,1)):
#        cdf[ti] = np.sum(pdf[:ti])
        if pdf[ti] > 0:
            li = pdf[ti]
            while li >= res_bin:
                tx[lbin] += res_bin * (ti + tmin)
                li -= res_bin
                lbin += 1; res_bin = 1
                lbin = min(lbin, nbins-1)
            # and save the leftovers
            res_bin -= li
            tx[lbin] += li * (ti + tmin)
       
#    # Last sort into symmetric profile with tmax in middle
#    tx_profile = np.zeros(nbins)
#    tx_profile[:nbins//2] = tx[range(0, nbins-1,2)]
#    tx_profile[:nbins//2-1:-1] = tx[range(1, nbins, 2)]
    tx_profile = tx
    
    # iterate on least-squares residual for discretization
    fit = res(tx_profile, pxl_lsq, model)
    res_lsq = norm(fit - meas)
    
    return tx_profile, res_lsq, pxl_lsq

def res(tx, pxl, model, tmin=300):
    '''
    Calculate linestrength fit model
    '''
    fit = np.zeros(np.shape(model)[0])
    nbins = np.shape(tx)[0]
    for t in tx:
        fit += pxl / nbins * model[:,int(t-tmin)]
    return fit

# The workhorse function
def tikhonov_sa(tx, pxl, meas, model, reg, eps = 1e-5, tmin = 300, nbins = 30):
    '''
    Simulated annealing algorithm based on Corana
    
    Minimize  ||     residual         || + || regularization ||
              ||meas - pxl * model(tx)|| + reg ||d2(tx) / dx2||
    
    Minimization function is nonlinear, so rather than separate objective function,
    just update the relevant portion for each search step.
    pxl * || meas/pxl - model(tx) ||
    
    Tends to produce uniform or parabolic profiles.    
    
    INPUS:
        tx = initial guess T(x) temperatures
        pxl = column density scaling factor, initial guess
        tmin = temperature of first
        meas = normalized linestrengths
        model = inversion matrix. Each column is temperature, 
                 starting at tmin and incrementing in 1 Kelvin
        AND simulated annealing parameters
        reg = regularization parameter
        t_sa = simulated annealing temperature
    TODO:
        lever-rule for Metropolis-Hastings search
        debugging
    '''
    ## Initialize Tikhonov fit residual
    tmin = int(tmin)
    tx = tx.astype(int)
    tx_best = tx.copy()
    pxl_best = pxl
    # Make L2 norm matrix
    l2 = 2 * np.diag(np.ones(nbins))
    l2 -= np.diag(np.ones(nbins-1),1) + np.diag(np.ones(nbins-1),-1)
    l2[0,:] = 0; l2[-1,:] = 0 # ignore edges
    l2[-1,-1] = 2; l2[-1,-2] = -2
    # evaluate vector of fit residual = measurement - model
    res = meas / pxl
    for ti in tx:
        res -= model[:, ti - tmin] / nbins
    fx_i = norm(res) * pxl + reg * norm(l2 @ tx)
    fx_best = fx_i
    
    ## initialize simulated annealing parameters from Corana
    # simulaed annealing steps
    t_sa = 10 # simulated annealing temperature
    cool_rate = 0.85
    n_t = min(100, 5 * nbins)
    return_flag = 0
    cooling_steps = 0 # number of times cooled
    max_steps = 100
    # metropolis-hastings solution search parameters
    lb = tmin * np.ones(nbins+1)
    ub = lb + np.size(model,1) - 1
    tstep = 50 * np.random.rand(nbins+1)
    minstep = 2 * np.ones(nbins+1)
    maxstep =  200 * np.ones(nbins+1)
    dilate = 1.8 * np.ones(nbins + 1) # how much to change the step size
    n_s = 20
    n_accept = np.zeros(nbins + 1)
    # Added Malarich parameters for tweaking Metropolis-Hastings search
    lever = 0.2
    w_ub = 0.01
    # special search parameters for pxl_factor
    lb[-1] = pxl / 1.5
    ub[-1] = pxl * 1.5
    tstep[-1] = .05
    minstep[-1] = .005
    maxstep[-1] = (ub[-1] - lb[-1]) / 5
    
    # Begin iterative optimization algorithm
    while return_flag < 1 and cooling_steps < max_steps:
        fx_start = fx_best
        for m in range(n_t):
            # update step size n_t times at each annealing temperature
            for j in range(n_s):
                # run n_s times at each search step size
                for h in range(nbins):
    # Step 1: generate a new test point
                    keep_searching = True; n_search = 0
                    tx_old = tx[h]
                    while keep_searching:
                        tshift = int(tstep[h] * np.random.randn()) # Gaussian search
                        # and add non-Gaussian search
                        if h == 0:
                            tx_reg = .5 * (lb[h] + tx[h + 1])
                        elif h == (nbins-1):
#                            tx_reg = 0.5 * (lb[h] + tx[h - 1])
                            tx_reg = w_ub * ub[h] + (1 - w_ub) * tx[h-1]
#                        elif h == nbins // 2:
#                        # If you set tx_reg = -100, then result will be M-shape profile
#                            tx_reg = w_ub * ub[h] + (1 - w_ub)/2 * (tx[h+1] + tx[h-1])
                        else:
                            tx_reg = .5 * (tx[h+1] + tx[h-1])
                        tx_test = tx_old + tshift + int(lever * (tx_reg - tx_old))
    # Step 2: is x in a priori domain?                        
                        if tx_test >= lb[h] and tx_test <= ub[h]:
                            keep_searching = False
                        n_search += 1
                        if n_search > 20:
                            tstep[h] = min(tstep[h] + 10, maxstep[h])   
    # Step 3: evaluate x_test, accept test point?
                    tx[h] = tx_test
                    res -= (model[:,tx_test-tmin] - model[:,tx_old - tmin])/nbins
                    ftest = norm(res) * pxl + reg * norm(l2 @ tx)
                    if ftest <= fx_i: # automatically accept new point
                        fx_i = ftest
                        n_accept[h] += 1
                        if ftest < fx_best:
                            tx_best = tx.copy()
                            fx_best = ftest
                            pxl_best = pxl
                    else: # possibly accept new point using annealing algorithm
                        probability_accept = np.exp((fx_i - ftest) / t_sa)
                        if probability_accept > np.random.rand():
                            fx_i = ftest
                            n_accept[h] += 1
                        else: # reject test point
                            tx[h] = tx_old
                            res += (model[:,tx_test-tmin] - model[:,tx_old - tmin])/nbins
    # Repeat steps 1-3 for pxl scaling parameter
                h = nbins
                keep_searching = True; n_search = 0
                while keep_searching:
                    pxl_test = pxl + (-1 + 2 * np.random.randn()) * tstep[-1]
                    n_search += 1
                    if pxl_test >= lb[h] and pxl_test <= ub[h]:
                        keep_searching = False
                    if n_search > 20:
                        tstep[h] = min(tstep[h] * 2, maxstep[h])   
                res += meas * (1/pxl_test - 1/pxl)
                ftest = norm(res) * pxl_test + reg * norm(l2 @ tx)
                if ftest <= fx_i: # automatically accept new point
                    fx_i = ftest
                    n_accept[h] += 1
                    pxl = pxl_test
                    if ftest < fx_best:
                        tx_best = tx.copy()
                        fx_best = ftest
                        pxl_best = pxl_test
                else: # possibly accept new point using annealing algorithm
                    probability_accept = np.exp((fx_i - ftest) / t_sa)
                    if probability_accept > np.random.rand():
                        fx_i = ftest
                        n_accept[h] += 1
                        pxl = pxl_test
                    else: # reject test point
                        res -= meas * (1/pxl_test - 1/pxl)
    # Step 5: update simulated annealing search parameters
            for u in range(nbins + 1):
                if n_accept[u] > 0.6 * n_s: # accepted >60% test points
                    step_new = tstep[u] * (1 + dilate[u] * (n_accept[u]/n_s - .6)/.4)
                    tstep[u] = min(maxstep[u], step_new)
                elif n_accept[u] < 0.4 * n_s: # didn't accept enough points
                    step_new = tstep[u] / (1 + dilate[u] * (.4 - n_accept[u]/n_s)/.4)
                    tstep[u] = max(minstep[u], step_new)
            n_accept[:] = 0 # reset count
    # Step 6: reduce simulated annealing "temperature"
        cooling_steps += 1
        t_sa *= cool_rate    
    # Step 7: does latest step 1-6 loop satisfy exit conditions?
        # exit conditions here: two cooling steps of miniscule residual improvement
        if fx_start - fx_best < eps:
            return_flag += 0.5
            # extra tweak: don't want to exit if t_sa too high and can't find minima
            if np.average(tstep[:-1]) > .99 * maxstep[0]:
                t_sa *= .2
                return_flag = 0
        else:
            return_flag = 0
        tx = tx_best.copy()
        pxl = pxl_best
        fx_i = fx_best
        # Last calcualte linestrength fit residual
        res = meas / pxl
        for t in tx:
            res -= model[:, t - tmin] / nbins
        res_best = pxl * norm(res)
                    
    return tx_best, pxl_best, res_best

'''
Other duplicated functions if you don't want object-oriented framework
'''
def lbin_sa(meas, model, start = 'uniform', nbins = 30, 
            reg_weight = 10.**np.asarray([-2,-1,0,1,2])):
    '''
    Length-bin simulated annealing master code.
    
    The non-object-oriented version of Snorm2Tx.length_bin()
    
    INPUTS:
        meas = normalized linestrengths vector
        model = PDF vs normalized linestrengths
        nbins = number of length-bins in temperature profile solution
        reg_weight = array of 
    '''
    
    tic = time()
    # Make L2 norm matrix
    l2 = 2 * np.diag(np.ones(nbins))
    l2 -= np.diag(np.ones(nbins-1),1) + np.diag(np.ones(nbins-1),-1)
    l2[0,:] = 0; l2[-1,:] = 0 # ignore edges
    # calculate the uniform and least-squares cases
    t_uni, res_uni, pxl_uni = uniform_fit(meas, model)
    reg_uni = 0 # by definition
    t_lsq, res_lsq, pxl_lsq = lsq_fit(meas, model)
    reg_lsq = norm(l2 @ t_lsq)
    reg_weight = np.asarray(reg_weight) * res_lsq
    if start == 'uniform':
        tx0 = t_uni
        pxl0 = pxl_uni
    else:
        tx0 = t_lsq
        pxl0 = pxl_lsq
    # run the Tikhonov
    tx_ = np.ones((nbins, len(reg_weight)))
    pxl_ = np.ones(len(reg_weight))
    res_ = np.ones(len(reg_weight))
    reg_ = np.ones(len(reg_weight))
    for i, reg in enumerate(reg_weight):
        tx_[:,i], pxl_[i], res_[i] = tikhonov_sa(tx0, pxl0, meas, model, reg,
              eps = res_lsq / 100, nbins = nbins)
        reg_ = norm(l2 @ tx_[:,i])
    # Now calculate the L-curve
    res_lsq = min(res_lsq, min(res_)) # mitigate discretization error
    reg_rectangle = np.sqrt( ((reg_ - reg_uni) / reg_lsq)**2 + 
                            ((res_ - res_lsq) / res_uni)**2)
    corner = np.argmin(reg_rectangle)
    print('Regularization in %d seconds' % (time() - tic))
    
    return tx_, res_, pxl_, reg_, corner

def objective(meas, tx, pxl, model, tmin=300):
    '''
    Calculate components of the objective function that tikhonov_sa() minimizes.
    '''
    # linestrength residual
    res = meas / pxl
    nbins = len(tx)
    for t in tx:
        res -= model[:, t - tmin] / nbins
    res_norm = norm(res) * pxl
    
    # and regularization
    l2 = 2 * np.diag(np.ones(nbins))
    l2 -= np.diag(np.ones(nbins-1),1) + np.diag(np.ones(nbins-1),-1)
    l2[0,:] = 0; l2[-1,:] = 0 # ignore edges
    reg_norm = norm(l2 @ tx)
    
    return res_norm, reg_norm


# -*- coding: utf-8 -*-
"""
Support scripts for Ebinning

Created on Mon Jul 22 12:18:12 2019

@author: Nate the Average
"""
import numpy as np

# hapi

try:
    from ..hapi import HITRAN_DEFAULT_HEADER, partitionSum 
    from ..hapi import absorptionCoefficient_Voigt, absorptionCoefficient_SDVoigt
except:
    from hapi_combustion import HITRAN_DEFAULT_HEADER, partitionSum, absorptionCoefficient_Voigt
    from hapi_combustion import absorptionCoefficient_SDVoigt
#find_package('pldspectrapy')
#from pldspectrapy.pldhapi import absorptionCoefficient_Voigt
HITRAN_DEFAULT_HEADER['default']['line_mixing_flag'] = 'E'
HITRAN_DEFAULT_HEADER['default']['gp'] = 0
HITRAN_DEFAULT_HEADER['default']['gpp'] = 0
HC_K = 1.4388028496642257 # physical constants for linestrength equation

###################
# S calculations
##################
def calc_smax(S0, elower, Tmax_gas):
    """
    Calculates S(T) where dS/dT = 0.
    
    Nest function for makeSrat() ReThrsh, where want to know if line should float or is too weak.
    
    TODO:
        Think about where I need this function.
        Maybe I should set up Smax just once when I initialize lineDict.
        And then think about precisely which parameters I need in lineDict vs sticking straight in .ref file.
    
    """
    S0 = float(S0); elower = float(elower) # make sure any strings converted to float.
    powerLaw = 2.77 # Power-law fit to T*Q(T) for H2O
    Tmax_S = min(HC_K * elower / powerLaw,Tmax_gas) #Cut off super highT lines if don't expect any temperature above 1300K
    if(Tmax_S > 296): #Don't expect sub-room temperature lines
        Smax = S0 * (Tmax_S / 296.)**(-powerLaw) * np.exp(-HC_K*elower*(1/Tmax_S - 1/296.))
    else:
        Smax = S0 # for all those lines with small E"
        
    return Smax


def snorm_fit_t_x(vars, s_e_array, thermo_old):#T_out, x_out, S0, elower, t_celsius, chi):
    """
    Calculates integrated area.
    Function to be called into scipy.optimize for updating T,X from initial guess

    find t_new, x_new to minimize integrated area residual
    min || int_alpha - S(T_new) P x_new L ||
    knowing snorm_meas = int_alpha / (S(T_0) P x_0 L)
    and assuming you can ignore scaling factors in front of equation
    which might be okay when already weighting each E"-bin by snorm_uc, yields
    
    min || snorm_meas (S(T_0) x_0 ) / (S(t_new) x_new) - 1 ||
    
    Uses power-law approximation for partition function Q_H2O(T)
    
    INPUTS:
        s_e_array: 3-column numpy float, with [E", S_norm, UC]
                 S_norm = (integrated area) / (S(T_old) P X_old L)
         thermo_old: [T_old (Celsius), x_old (molefraction)]
    """
    t_old = thermo_old[0] + 273 # initial guess temperature (Kelvin)
    x_old = thermo_old[1]
    
    thermo_old = vars.copy()
    
    # Fitted variables
    t_new = vars[0] + 273 # temperature (Kelvin)
    x_new = vars[1] # molefraction
    
    elower = s_e_array[:,0]
    snorm = s_e_array[:,1]
    uc = s_e_array[:,2]
    
    powerLaw = 2.77
    st_new_old = (t_new / t_old)**(-powerLaw) * np.exp(-HC_K * elower * (1/t_new - 1/t_old)) # S(T)_new / S(T)_old
    xt_old_over_new = (x_old / x_new) / st_new_old
    
    return (snorm * xt_old_over_new - 1) / uc # (nE x 1) array to minimize

def s_bkgd(vars, s_e_array, t_celsius, chi):
    '''
    Fits one temperature and roomT column density
    '''
    temp_old = t_celsius + 273
    t_new = vars[0]
    x_new = vars[1]
    x_bkgd = vars[2]
    
    elower = s_e_array[:,0]
    s_norm = s_e_array[:,1]
    uc = s_e_array[:,2]

    power = -2.77
    s_meas = s_norm * chi * temp_old**power * np.exp(-HC_K * elower / temp_old)
    s_t = t_new**power * np.exp(-HC_K * elower / t_new) * x_new
    s_bkgd = 296**power * np.exp(-HC_K * elower / 296) * x_bkgd
    
    return (s_meas - s_t - s_bkgd) / uc

def s_two_temp(vars, s_e_array, t_celsius, chi):
    '''
    Fits (T1, x1, T2, x2) from linestrength curve.
    Output temperatures in Kelvin
    '''
    temp_old = t_celsius + 273
    t1 = vars[0]
    t2 = vars[1]
    x1 = vars[2]
    x2 = vars[3]
    elower = s_e_array[:,0]
    s_norm = s_e_array[:,1]
    uc = s_e_array[:,2]
    
    power = -2.77
    s_meas = s_norm * chi * temp_old**power * np.exp(-HC_K * elower / temp_old)
    s_t1 = t1**power * np.exp(-HC_K * elower / t1) * x1
    s_t2 = t2**power * np.exp(-HC_K * elower / t2) * x2    
    return (s_meas - s_t1 - s_t2) / uc

def s_t(t_new, t0, elower, power=-2.77):
    '''
    Returns linestrengths at s(t_new) given array elower
    '''
    return (t_new / t0)**power * np.exp(-HC_K * elower * (1/t_new - 1/t0))

def s_t_exact(t_new, t0, elower, molec_id = 1, iso = 1):
    '''
    Returns linestrengths at s(t_new) given array elower and partition function q
    INPUTS:
        t_new = output evaluated temperature (Kelvin)
        t0 = reference temperature, 296 for HITRAN (Kelvin)
        elower = E" at which to calculate S(E") (int, float, numpy array)
        molec_id, iso = HITRAN molecule reference codes for partition function
    '''
    q_t0 = partitionSum(molec_id, iso, t0)
    if type(t_new) is np.ndarray:
        q_tnew = np.asarray(partitionSum(molec_id, iso, list(t_new)))
    else:
        q_tnew = np.asarray(partitionSum(molec_id, iso, t_new))
    boltz_t0 = np.exp(-HC_K * elower / t0) / (q_t0 * t0)
    boltz_t_new = np.exp(-HC_K * elower / t_new) / (q_tnew  * t_new)
    return boltz_t_new / boltz_t0

#####################
# Write par file ####
#####################
    
def write_par_line(line_dict, out_file):
    '''
    Write 160-char Hitran format transition in line_dict to out_file. 
    '''
    added_quanta = False
    for name in HITRAN_DEFAULT_HEADER['order']:
        try:
            value = line_dict[name]
        except KeyError:
            value = HITRAN_DEFAULT_HEADER['default'][name]
        if 'quanta' in name:
            # only count the quanta once
            if added_quanta is False:
                out_file.write('%60s' % line_dict['quanta'][:60])
                added_quanta = True
        else:
            name_format = HITRAN_DEFAULT_HEADER['format'][name]
            try:
                str_len = int(name_format.split('.')[0][1:])
            except ValueError:
                str_len = int(name_format[1:-1])
            if 's' not in name_format:
                value = float(value)
            value_str = name_format % value
            # Fix leading-zero bug on air-broadening parameters
            if len(value_str) > str_len:
                if value >= 0:
                    out_file.write(value_str[1:])
                else:
                    out_file.write('-.' + value_str.split('.')[1])
            else:
                out_file.write(value_str)
    out_file.write('\n')

############################
## Simulate absorption spectra (TD for time-domain, FD for frequency-domain)
############################
def calc_spectrum_fd(xx, mol_id, iso, molefraction, pressure, temperature, 
                     pathlength, shift, db_file_name='H2O', sdvoigt = False):
    '''
    Spectrum calculation for adding multiple models with composite model.
    
    See lmfit model page on prefix, parameter hints, composite models.
    
    INPUTS:
        xx -> wavenumber array (cm-1)
        name -> name of file (no extension) to pull linelist
        mol_id -> Hitran integer for molecule
        iso -> Hitran integer for isotope
        molefraction
        pressure -> (atmospheres)
        temperature -> kelvin
        pathlength (centimeters)
        shift -> (cm-1) calculation relative to Hitran
        db_file_name -> string name (no extension) of Hitran linelist file
    
    TODO: how to pull molecule name from prefix of call?
    '''
    if sdvoigt:
        nu, coef = absorptionCoefficient_SDVoigt(((int(mol_id), int(iso), molefraction),),
                db_file_name, HITRAN_units=False,
                OmegaGrid = xx + shift,
                Environment={'p':pressure,'T':temperature},
                Diluent={'self':molefraction,'air':(1-molefraction)}) 
    else:
        nu, coef = absorptionCoefficient_Voigt(((int(mol_id), int(iso), molefraction),),
                db_file_name, HITRAN_units=False,
                OmegaGrid = xx + shift,
                Environment={'p':pressure,'T':temperature},
                Diluent={'self':molefraction,'air':(1-molefraction)})
    absorp = coef * pathlength
    return absorp       
            
def calc_cepstrum_td(xx, mol_id, iso, molefraction, pressure, temperature, 
                     pathlength, shift, db_file_name='H2O', sdvoigt = False):
    '''
    Spectrum calculation for adding multiple models with composite model.
    
    See lmfit model page on prefix, parameter hints, composite models.
    
    INPUTS:
        xx -> wavenumber array (cm-1)
        name -> name of file (no extension) to pull linelist
        mol_id -> Hitran integer for molecule
        iso -> Hitran integer for isotope
        molefraction
        pressure -> (atmospheres)
        temperature -> kelvin
        pathlength (centimeters)
        shift -> (cm-1) calculation relative to Hitran
        db_file_name -> string name (no extension) of Hitran linelist file
    
    TODO: how to pull molecule name from prefix of call?
    '''
    absorp_fd = calc_spectrum_fd(xx, mol_id, iso, molefraction, pressure,
                                 temperature, pathlength, shift, db_file_name,
                                 sdvoigt)
    absorp_time_domain = np.fft.irfft(absorp_fd)
    return absorp_time_domain        

def spectra_allTD(vars, xx, data):
    molec_id = vars[0]
    local_iso_id = vars[1]
    pressure_atm = vars[2]
    temperature_kelvin = vars[3]
    molefraction = vars[4]
    pathlengths = vars[5]
    shift = vars[6]
    linelists = vars[7]
    absorption_all = np.zeros(len(xx))
    for i in range(len(pathlengths)):
        nu, coef = absorptionCoefficient_Voigt(((int(molec_id), int(local_iso_id), molefraction),),
                linelists[i], HITRAN_units=False,
                OmegaGrid = xx + shift,
                Environment={'p':pressure_atm,'T':temperature_kelvin},
                Diluent={'self':molefraction,'air':(1-molefraction)})
        absorption_all += pathlengths[i] * coef
    absorp_td = np.fft.irfft(absorption_all)
    return data - absorp_td

###############################
# Simulate nonuniform-path absorption spectra (frequency-domain)
###############################



from setuptools import setup
import codecs
from os import path

def read(rel_path):
    here = path.abspath(path.dirname(__file__))
    with codecs.open(path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

# Get the long description from the relevant files
readme = read('README.md')
changelog = read('CHANGELOG.md')
desc = readme + '\n\n' + changelog

setup(
      name="ntfit",
      version=get_version("ntfit/__init__.py"),
      description="nonuniform-temperature fitting",
      url="https://gitlab.com/nam5312/ntfit",
      author="Nathan Malarich",
      author_email="nathan.malarich@colorado.edu",
      packages=['ntfit','ntfit.hapi','ntfit.spectrafit'],
      install_requires=[ # combustion-custom Hapi included in package
          "numpy", # standard math/arrays, also numpy.linalg.norm and numpy.matlib.repmat
          "matplotlib",
          "scipy", # for nonlinear least-squares fitting
      ],
      long_description = desc,
      long_description_content_type = 'text/markdown',

    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',  
        'Operating System :: Microsoft :: Windows :: Windows 10',      
        'Programming Language :: Python :: 3.6',
    ],
)